/***********************************************
************************************************
**                                            **
** mamot                                      **
** a program for HMM modelling                **
** 2008, Mauro C. Delorenzi                   **
**                                            **
************************************************
************************************************

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

extern class ModelFB model;
extern bool  bfileoutput, bdsDNA, bdsDNAjointly;
extern int nbseqs, maxlen;

#include "PlatformMain.h"

extern int 	seqNb, seqLen;
extern int  seq[kMaxSeqLen+1];
extern bool completedFile;
extern char	SeqName[kMaxSeqName+2];
extern char *extPFile;
extern FILE *cFpWritePP, *cFpWritePPD,*cFpWriteCP; 
extern FILE *cFpWriteDom; 
extern FILE *cFpReadSeq; 
extern bool verbose, bUseExtP, FBDdetailsD;
FILE *cFpWriteModel;

// ---------------------------------------------------------------------------
//		 PlatformMain 					
// ---------------------------------------------------------------------------

void PlatformMain(const char *ModelFile, char *seqFile, int modus, int n )
{
FILE  *fprm;

if (verbose) cerr << "PlatformMain, modus = " << modus << endl;
if  ((fprm = fopen( ModelFile,"r" ))  == NULL )
	{cerr << "\n could not open the model filecalled: "<< ModelFile<< "\n"<< endl; exit(1);}
else 
	{ 
	if (! ReadModel(fprm, modus)) {cerr << "\n ERROR in ReadModel \n" << endl; exit(1);}
	}  	
  
if (verbose)  cerr << "\nFinished ReadModel";
fclose( fprm );

if (bUseExtP)
{
	if (verbose) cerr << "\nStarting ReadExtP";
	{
	if  ((fprm = fopen( extPFile,"r" ))  == NULL )
		{ cerr << "\n could not open the external prob file called: " << extPFile << "\n"<< endl; exit(1); }
	else 
		{ 
		if (model.ReadextP(fprm) ) {cerr << "\n ERROR in ReadextP \n" << endl; exit(1);}
		}  	
	if (verbose)  cerr << "\nfinished ReadExtP";
	fclose( fprm );
}}

if (verbose) cerr << "\nswitch-modus = " << modus << endl;

 switch  (modus)
	{
	  //FB-probability
	case  0 : FBmain(ModelFile, seqFile);     
	  if (verbose)  cerr  <<  "FBmain=>PlatformMain" <<  endl; 
	  break;   
	  //generation
	case  1 : GenMain(n, maxlen);	       
	  break;
	  //baum-welch	
	case  2 : if ((bdsDNA) || (bdsDNAjointly)) BWmaindsDNA(ModelFile, seqFile, nbseqs, maxlen); 
	  else BWmain(ModelFile, seqFile, nbseqs, maxlen); 
	  break;
	  //posterior-decoding	
	case  3 : bfileoutput = true; FBmain(ModelFile, seqFile); 	  
	  break;
	  //viterbi-Probability	//viterbi-decoding	
	case  4 : Vitmain(ModelFile, seqFile); 
	  break;
	  //viterbi-learning
	case  5 : if ((bdsDNA) || (bdsDNAjointly))
	  {VitLearnMaindsDNA(ModelFile, seqFile, nbseqs, maxlen);}
	  else {VitLearnMain(ModelFile, seqFile, nbseqs, maxlen);}
	  break;
	default: cerr << "\ndefault: Posterior Decoding\n"; 
		bfileoutput = true; FBmain(ModelFile, seqFile);      
	  break;
	}
}

/* PlatformMain
************************************/

// ---------------------------------------------------------------------------
// ReadModel 
// ---------------------------------------------------------------------------
bool ReadModel (FILE *cFrMod, int modus)
{
int nbletters=-1,  // -1 for "undefined"
nbstates=0,
nbEMties=0,
nbcEMties=0,
maxnbtiesingroup=0,
tiescounter;
bool error=false;
char buffer[kMaxLineLen];
char *tok, *tok1;


if (verbose)  cerr << "\nstarting ReadModel" ;

  //1.nb states  & alphabet
  rewind (cFrMod);
  
  while (fgets(buffer,kMaxLineLen,cFrMod)!= NULL)
    {
      if((strncmp(buffer,"State",5) == 0) ) nbstates++;
      // accept complementary pairs of ties: count how many for memory assignement
      if((strncmp(buffer,"compl_em",8) == 0) ) nbcEMties++;
      // accept ties
      if((strncmp(buffer,"tie_em",6) == 0) ) 
	{
	  nbEMties++; 	tiescounter=1;	
	  tok1 = strtok(buffer,"\t ");
	  while (tok1)
	    {tiescounter ++; tok1 = strtok(NULL,"\t \n");}
	  if (tiescounter > maxnbtiesingroup)  {maxnbtiesingroup = tiescounter;}
	}

      if((strncmp(buffer,"Alphabet",8) == 0) )
	{
	  if (nbletters!=-1) {
	    cerr << "Alphabet defined twice !" <<endl;
	    error=true;
	  }

	  tok = strtok(buffer, "\t\n ");
	  tok = strtok(NULL, "\t\n ");
	  nbletters = (tok==NULL ? 0 : strlen(tok));
	  if (verbose) cerr << "\n alphabet with " << nbletters << " letters" ;
	}
    }
   if (nbletters==-1) 
   		{
      	cerr  << "\n No alphabet defined." << endl;
      	error=true;
		}

  if (verbose) 
  	{
    cerr  << "\n After reading model, before memory allocation: ";
	cerr  << "\n Alphabet size = " << nbletters;
    cerr  << "\n maxnbtiesingroup = " << maxnbtiesingroup;	
    cerr  << "\n nbstates = " << nbstates;
    cerr  << "\n tie_em = " << nbEMties;
    cerr  << "\n compl_em = " << nbcEMties;
    cerr  << "\n call AssignMemory" << endl;
    }

  // If we found a problem in the file, abort
  if (error) return(!error);

  model.AssignMemory(modus, nbstates,nbletters,kMaxSeqLen,kMaxNameLen,nbEMties, maxnbtiesingroup, nbcEMties);

  model.usemodus = modus;
  if (verbose) cerr  << "\n model.usemodus = " << model.usemodus << endl;
  if (verbose) cerr  << "\n call ReadModelFB" << endl;
  
  cFpWriteModel = NULL;
  //re-read the model file and assign the parameters
  if (! (model.ReadModelFB(cFrMod, cFpWriteModel)))
    {
      model.Pseudocounts();
      model.InitCodes();
    }
  else {error=true;}
  
  return (!error);
  }
/* ReadModel
************************************/
//////////////////////////////////////////////////////////////////////////////////////////////////
