/***********************************************
************************************************
**                                            **
** mamot                                      **
** a program for HMM modelling                **
** 2008, Mauro C. Delorenzi                   **
**                                            **
************************************************
************************************************

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Vitmain.h"
extern class ModelFB model;
extern bool verbose, bstateweights, bUseExtP;

extern  int   seqNb, seqLen;
extern  char  SeqName[kMaxSeqName+2];
extern  int   seq[kMaxSeqLen+1];
extern  FILE *cFpReadSeq;
FILE *cFpWriteVIT, *cFpWriteVITDEC;

extern bool   bdsDNA;

// ---------------------------------------------------------------------------
//		 Vitmain 				
// ---------------------------------------------------------------------------
void Vitmain(const char *ModelFile, char *seqFile )
{
bool completedFile;
int seqNb;
char *tok;
if (verbose) {cerr << "\nVitmain" <<endl;}

model.SetUpLogParameters();
if (bstateweights) model.SetUpLogParametersW();
model.SetUpConnections();

if  ((cFpWriteVIT = fopen( "Vitprob.txt", "w" ))  == NULL )
	{ cerr << "\n could not open the file  Vitprob for writing  \n"<<endl; exit(1);}

if  ((cFpWriteVITDEC = fopen( "Vitdecoding.txt", "w" ))  == NULL )
	{ cerr << "\n could not open the file Vitdecoding for writing\n"<<endl; exit(1);}

if  ((cFpReadSeq = fopen( seqFile, "r" ))  == NULL )
	{cerr << "\n could not open the file for the sequences with address  " << seqFile << "\n"; exit(1);}

//model.InitialiseViterbi();	
seqNb = 0;
completedFile = false;
while (completedFile == false) 
	{
	seqNb ++;	
	if ( !(ReadSeq (seqNb, seqLen, seq, completedFile, SeqName)) )
		{
		if (seqLen > 0)	{ WriteWarning(seqNb,1);}	
		else	{seqNb--;}
		} 
	if (seqLen > 0)
		{ 
		tok = strtok(SeqName, "\t  \n");
		fprintf(cFpWriteVIT, "\n%s",tok);
	  if (verbose)   cerr << ">>read seq-Nb " << seqNb << "  name = " << SeqName << endl;
	  if (verbose)   cerr << "processing seq-Nb " << seqNb << "  name = " << SeqName << endl;
	  if (bUseExtP)
	  	{
         if (seqLen > kMaxExtPLen)
{cerr<<"\n Sequence Length excessive for use of external prob matrix, limit= "<<kMaxExtPLen<<endl;}
		 else 
		    {
		    model.ViterbiExtP(seqNb, seqLen, seq, cFpWriteVITDEC);
			fprintf(cFpWriteVIT, " #ln(Vit-P)=%g ", (double) model.ViterbiPathLogProb);
			tok = strtok(NULL, "");
			}
		completedFile = true;// only one sequence with bUseExtP
		}
	  else 
	  	{
		model.Viterbi(seqNb, seqLen, seq, cFpWriteVITDEC);
		fprintf(cFpWriteVIT, " #ln(Vit-P)=%g ", (double) model.ViterbiPathLogProb);
		tok = strtok(NULL, "");
		if (bdsDNA) 
			{
			model.rcseq(seqLen, seq);
			seqNb +=1;
			fprintf(cFpWriteVIT, "\n%sSecondStrand",tok);
			model.Viterbi(seqNb, seqLen, model.SecondStrand, cFpWriteVITDEC);
			fprintf(cFpWriteVIT, " #ln(Vit-P)SecondStrand=%g ", (double) model.ViterbiPathLogProb);
			}
		}
	if ( (seqNb % kperiodStdout1) == 0)	 
		{
		if (seqNb == kperiodStdout1) cerr<<"One point "<<kperiodStdout1<<" sequences, one line "<<kperiodStdout2<<"\n"; 
		cerr<<".";  //fflush (stdout);
		}
	if ( (seqNb % kperiodStdout2) == 0)  cerr<<"\n";  	 
	if ( (seqNb % kperiodStdout3) == 0)  cerr<<"\n processed seq-nb= "<<seqNb<<"\n";
	}	}
cerr<<"\n";
fclose(cFpReadSeq);
fprintf (cFpWriteVIT,"\n");fclose(cFpWriteVIT);
fprintf (cFpWriteVITDEC,"\n");fclose(cFpWriteVITDEC);
}
/* Vitmain
************************************/
/////////////////////////////////////////////////////////////////////////////////////////////////////////
