/***********************************************
************************************************
**                                            **
** mamot                                      **
** a program for HMM modelling                **
** 2008, Mauro C. Delorenzi                   **
**                                            **
************************************************
************************************************

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

extern double CutoffForWriting;
extern int    Code[], Decode[];
extern int    CutoffForConnecting;
extern bool   verbose, writeControls, bfileoutput, FBDdetailsC;
extern long   seedran2;
extern bool   bupdateemiss, bupdatetrans, bdsDNA, bdsDNAjointly, buseties;
extern bool   bstateweights, busecties, bUseExtP;
extern int    seqNb;
extern double pseudocountweight, vpseudoTrans;
extern float  vMINProbPrint;
extern char   *extPFile;

#include "classModelFB.h"

//class implementation

// ---------------------------------------------------------------------------
//	 ReadModelFB   //read model from file
// ---------------------------------------------------------------------------
bool ModelFB::ReadModelFB (FILE *cFrMod,FILE *cFpWriteModel ) 
{
  int   k, s, t, n;
long double sum;
  unsigned int len;
  float number=0;
  bool  error=false;
  char	buffer[kMaxLineLen];
  char	*tok1,*tok2,*tok;
  char	name[kMaxNameLen];
  int tiescounter;
#define SEPCHARS " \t\n"

  if (verbose)
    {
      cerr << "\nstarting ModelFB::ReadModelFB, with:";
      cerr << "\n  nbLetters= " << nbLetters;
      cerr << "\n  nbStates= " << nbStates;
      cerr << "\n  maxSeqLength= " << maxSeqLength << endl;
    }

  // REWIND AND FIRST READ THE NAMES OF THE STATES AND THE ALPHABET BEING USED

  rewind (cFrMod);s=-1;
  while (fgets(buffer,kMaxLineLen,cFrMod)!= NULL)
    {

      if((strncmp(buffer,"Alphabet",8) == 0) )
	  {
	  tok = strtok(buffer,SEPCHARS);
	  tok = strtok(NULL, SEPCHARS);
	  for (t=0; t<nbLetters;  t++)   letter[t+1] = (char) tok[t];

	  if (verbose)
    	{
		for (t=0; t<nbLetters;  t++)   cerr << "\n tok[" << t << "] = " << (char) tok[t];
		for (t=1; t<=nbLetters;  t++)   cerr << "\n letter[" << t << "] = " << (char) letter[t];
		cerr << " "<< endl; 
    	}

		  if ((bdsDNA) || (busecties)|| (bdsDNAjointly)) 
	    	{ 
	    	  if  (nbLetters == 4)    
			{ 
			  if (!((letter[1] == 'A') && (letter[2] == 'C')&& (letter[3] == 'G')&& (letter[4] == 'T')))
		    	{ 
		    	cerr  << "\n with options -u, -v and -p  the Alphabet must be ACGT in this order"<< endl; 
		  		cerr  << "\n mamot did find in your model file  "<< nbLetters << " letters" << endl; 
			for (t=1; t<=nbLetters;  t++)   cerr << "\n letter[" << t << "] = " << (char) letter[t];
				cerr << " "<< endl; 
				  error=true;
		    	}
			}
	    	  else { 
		  		cerr  << "\n with options -u, -v and -p  the Alphabet must be of 4 letters ACGT in this order"<< endl; 
		  		cerr  << "\n mamot did find in your model file  "<< nbLetters << " letters" << endl; 
			for (t=1; t<=nbLetters;  t++)   cerr << "\n letter[" << t << "] = " << (char) letter[t];
				cerr << " "<< endl; 
		  		error=true;
		  		}
	    	}
		}

	
      if((strncmp(buffer,"State",5) == 0) )
		{
		  s++;
		  tok = strtok(buffer,SEPCHARS);
		  tok1 = strtok(NULL,SEPCHARS);
		  sscanf(tok1,"%s", name);//name of the state
		  len = strlen(name);
		  if (name[len-1] == (char) ':')  name[len-1]='\0';
		  strcpy(statenames[s], name);
		  if (strncmp(statenames[s],"END",3) == 0)   {endstate=s; }
		  if (strncmp(statenames[s],"BEGIN",5) == 0) {startstate=s;}
		}
    }

  // REWIND: NOW READ THE REST
  rewind (cFrMod);
  s=-1;nbEMties = 0;nbcEMties = 0;
  
  for (t=1; t<=nbLetters ; t++)	{nullEmFreq[t] = 0.0;}


  while (fgets(buffer,kMaxLineLen,cFrMod) != NULL)
    {      

      if (buffer[strlen(buffer)-1]=='\n') buffer[strlen(buffer)-1]='\0';// Remove trailing \n

      tok = strtok(buffer,SEPCHARS);

      if (tok==NULL) continue;

      if((strncmp(tok,"State",5) == 0)) 	{s++;}

 
      if((strncmp(tok,"E:",2) == 0))
	  {
	  for (t=1; t<=nbLetters ; t++)//read all emProbs of this state, assuming complete list
	    {
	      tok1 = strtok(NULL,SEPCHARS);
	      if (tok1)
		  {
		  number = (float) atof (tok1);
		  emProb[s][t] = number;
		  }
	      else //multiline E:
		  {
		  fgets (buffer,kMaxLineLen,cFrMod);
		  tok = strtok(buffer,SEPCHARS);
		  tok1 = strtok(NULL,SEPCHARS);
		  if (tok1)
		    {
		      number = (float) atof (tok1);
		      emProb[s][t] = number;
	  } }}	} 


      if((strncmp(tok,"T:",2) == 0))
	{
	  k=-1;	  //allow multiline here
	  tok1 = strtok(NULL,SEPCHARS);  
	  tok2 = strtok(NULL,SEPCHARS);  
	  number = (float) atof (tok2);

	  while (tok2)
	    {
	      number = (float) atof (tok2);
	      len = strlen(tok1);k=-1;
	      for (t=0; t<nbStates; t++) 
		{ 
		  if (len == strlen(statenames[t])) {  if ((strncmp(tok1,statenames[t],len) == 0)) k=t;}
		}
	      if (k==-1) 
		{
		  cerr<<"\n error: the state " << tok1 << " in the \"T:\" section of state "<<statenames[s]<<
		    "could not be identified as a valid state"; error=true;
		}
	      else trProb[s][k] = number;
	      tok1 = strtok(NULL,SEPCHARS);
	      tok2 = strtok(NULL,SEPCHARS);
	    }
	}	


      if((strncmp(tok,"NULLMODEL:",10) == 0))
	{
	  for (t=1; t<=nbLetters ; t++)
	    {
	      tok1 = strtok(NULL,SEPCHARS);
	      if (tok1)
		{
		  number = (float) atof (tok1);
		  nullEmFreq[t] = number;
		}
	      else //multiline
		{
		  fgets (buffer,kMaxLineLen,cFrMod);
		  tok = strtok(buffer,SEPCHARS);
		  tok1 = strtok(NULL,SEPCHARS);
		  if (tok1)
		    {
		      number = (float) atof (tok1);
		      nullEmFreq[t]   = number;
		    }	}	}	} 


      if((strncmp(tok,"fix_tr",6) == 0))
	{k=-1;
	  tok1 = strtok(NULL,SEPCHARS);
	  while (tok1)
	    {
	      len = strlen(tok1);k=-1;
	      for (t=0; t<nbStates; t++) 
		{ 
		  if (len == strlen(statenames[t])) { if ((strncmp(tok1,statenames[t],len) == 0)) k=t;}
		}
	      if (k==-1) 
		{
		  cerr<<"\n error: the state " << tok1 << " in the \"fixed trans:\" section could not be identified as a valid state"; 
		  error=true;
		}
	      else   updatetr[k]= 0;
	      tok1 = strtok(NULL,SEPCHARS);
	    }
	}


      if((strncmp(buffer,"tie_em",6) == 0) ) // accept ties: count how many for memory assignement
	{
	  nbEMties++; 	tiescounter =0;	
	  tok1 = strtok(NULL,"\t ");
	  while (tok1)
	    {
	      tiescounter ++; k=-1;
	      len = strlen(tok1);
	      for (t=0; t<nbStates; t++) 
		{ 
		  if (len == strlen(statenames[t])) { if ((strncmp(tok1,statenames[t],len) == 0)) k=t; }
		}
	      if (k==-1) 
		{
		  cerr<<"\n error: the state " << tok1 << " in a \"tie_em:\" line could not be identified as a valid state"; 
		  error=true;
		}
	      else   {ties[nbEMties][tiescounter]= k;    ties[nbEMties][0]= tiescounter;}
	      tok1 = strtok(NULL,SEPCHARS);
	    }
	}	


      if((strncmp(buffer,"weight_em",9) == 0) ) // give weights to some states ' emissions
	{
	      tok1 = strtok(NULL,SEPCHARS);
	      if (tok1)
		{ number = (float) atof (tok1);  
		  }
 	  tok1 = strtok(NULL,SEPCHARS);// read the states
	  while (tok1)
	    {
	      k=-1;
	      len = strlen(tok1);
 	      for (t=0; t<nbStates; t++) 
		{  // identify the state
		  if (len == strlen(statenames[t])) { if ((strncmp(tok1,statenames[t],len) == 0)) k=t; }
		}
	      if (k==-1) 
		{
		  cerr<<"\n error: the state " << tok1 << " in a \"weight_em:\" line could not be identified as a valid state"; 
		  error=true;
		}
	      else   {weightEM[k]= number;}
	      tok1 = strtok(NULL,SEPCHARS);
	    }
	}	


      if((strncmp(buffer,"compl_em",8) == 0) ) 
	{
	  nbcEMties++; n=-1;	
	  tok1 = strtok(NULL,"\t ");
	  while (tok1)
	    {
	      tiescounter ++; k=-1; n++;
	      len = strlen(tok1);
	      for (t=0; t<nbStates; t++) 
		{ 
		  if (len == strlen(statenames[t])) { if ((strncmp(tok1,statenames[t],len) == 0)) k=t; }
		}
	      if (k==-1) 
		{
		  cerr<<"\n error: the state " << tok1 << " in a \"compl_em:\" line could not be identified as a valid state"; 
		  error=true;
		}
	      else   
		{
		  if (n <= 1) {cties[nbcEMties][n]= k;}
		  else
		    {cerr<<"\n error: found more than two states  in a \"compl_em:\" line"; }
		}
	      tok1 = strtok(NULL,SEPCHARS);
	    }
	}	


    if((strncmp(tok,"fix_em",6) == 0))
	{k=-1;
	  tok1 = strtok(NULL,SEPCHARS);
	  while (tok1)
	    {
	      len = strlen(tok1);k=-1;
	      for (t=0; t<nbStates; t++) 
		{ 
		  if (len == strlen(statenames[t])){ if ((strncmp(tok1,statenames[t],len) == 0)) k=t;}
		}
	      if (k==-1) 
		{
		  cerr<<"\n error: the state " << tok1 << " in the \"fixed emiss:\" section could not be identified as a valid state"; 
		  error=true;
		}
	      else   updateem[k]= 0;
	      tok1 = strtok(NULL,SEPCHARS);
	    }
	}
    }
	
  updateem[endstate]= 0;
  updateem[startstate]= 0;
  pseudow[startstate]= 0;
  pseudow[endstate]= 0;

// check model for approximative sum of probs to 1, warning to stderr
for (s=0; s<nbStates;  s++)
	{
	sum=0;
	for (t=1; t<=nbLetters; t++) 
		{
		sum+= emProb[s][t];
		}
	if ((sum > ktol2 ) || (sum < ktol1 ))
		{	
		if ((strncmp(statenames[s],"BEGIN",5) != 0) && (strncmp(statenames[s],"END",3) != 0))
			{
			fprintf(stderr, "\nWarning: emission probabilities of state %s  ", statenames[s]);
			fprintf(stderr, "do not sum up to 1 but to  %2.9g \n", (double)sum);
		}}
	n=0;sum=0;
	for (t=0; t<nbStates; t++) 
		{
		if (trProb[s][t]>0)
			{
			n++;
			sum += trProb[s][t];
			}
		}
	if ((sum > ktol2 ) || (sum < ktol1 ))
		{
		if (strncmp(statenames[s],"END",3) != 0) 
			{
			fprintf(stderr, "\nWarning: transition probabilities of state %s  ", statenames[s]);
			fprintf(stderr, "do not sum up to 1 but to  %2.9g \n", (double)sum);
		}}
	}

  if (verbose) { cerr << "\n ending ModelFB::ReadModelFB"<< endl; }

return error;
}

/* ReadModelFB
************************************/


// ---------------------------------------------------------------------------
//	 ReadextP  //read additional scores from file 
// ---------------------------------------------------------------------------
bool ModelFB::ReadextP (FILE *cFrMod)  //read additional scores from file
{
int   k, s, t, pos, nbs=0;
int   ind[nbStates];
unsigned int len;
bool  error=false;
char  buffer[kMaxLineLen], line[kMaxLineLen];
char  *tok;
#define SEPCHARS " \t\n"

if (verbose){cerr << "\nstarting ModelFB::ReadextP" << endl;}

for (t=0; t<nbStates; t++) 
	{
	for (k=0; k<kMaxExtPLen; k++)  extP[t][k]= 1;
	}

while (fgets(buffer,kMaxLineLen,cFrMod)!= NULL)
    {
	strcpy(line, buffer);
    if((strncmp(buffer,"states:",7) == 0) )
	  {
	  s=0;
	  tok = strtok(buffer, SEPCHARS); 
	  tok = strtok(NULL, SEPCHARS);
	  while (tok)
	    {
	  	len = strlen(tok);k=-1;
	  	for (t=0; t<nbStates; t++) 
			{ 
			if (len == strlen(statenames[t])) 
				{
				if ((strncmp(tok,statenames[t],len) == 0)) 
					{
					k=t; 
	    			if (s >= nbStates) 
						{
						cerr<<"\n error: too many states " << tok << " in ReadextP:"; 
		  				error=true;
						}
					else 
						{
						ind[s]=t; nbs=s; 
						s=s+1;
						}
					}
			}	}
	    if (k==-1) 
			{
			cerr<<"warning: the state " << tok << " in ReadextP: could not be identified as a valid state"<< endl;; 
			}
	     tok = strtok(NULL, SEPCHARS);
	    }	
	}

	else
	  {
		tok = strtok(buffer, SEPCHARS);
		if (tok) 
		{
		  pos =  atoi (tok);// for the positions present in the file, need not be all
		  for (k=0; k<=nbs; k++) // for the list of states present in the file, need not be all
			{
	    	tok = strtok(NULL, SEPCHARS);
			if (!tok)
				{
				cerr<<"error: too few  number values in ReadextP: in line " << line << endl;
		  		error=true;
				}
			extP[ind[k]][pos]= (float) atof (tok); 
			}
	    }
	  }
	}
return error;
}

/* ReadextP
************************************/

// ---------------------------------------------------------------------------
//	 WriteModelFB  //write model to file
// ---------------------------------------------------------------------------
void  ModelFB::WriteModelFB(FILE *cFpWriteModel)  
{
if (verbose) { cerr << "\n starting ModelFB::WriteModelFB"; }
int s,t,n;
long double sum;

fprintf(cFpWriteModel, "Alphabet: ");
for (t=1; t<=nbLetters; t++) {fprintf(cFpWriteModel, "%c",letter[t]);}
if (verbose)
	{
	cerr << "\n Alphabet:";
	for (t=1; t<=nbLetters; t++) {cerr << "\t" << letter[t];}
	cerr << endl;
	}
	
fprintf(cFpWriteModel, "\n################################################");
for (s=0; s<nbStates;  s++)
	{
	fprintf(cFpWriteModel, "\nState %s  \n\tE: ", statenames[s]);
	sum=0;
	for (t=1; t<=nbLetters; t++) 
		{
		fprintf(cFpWriteModel, "  %g", (double)emProb[s][t]);
		sum+= emProb[s][t];
		}
	if ((sum > ktol2 ) || (sum < ktol1 ))
		{	
		if ((strncmp(statenames[s],"BEGIN",5) != 0) && (strncmp(statenames[s],"END",3) != 0))
			{
			fprintf(stderr, "\nWarning: emission probabilities of state %s  ", statenames[s]);
			fprintf(stderr, "do not sum up to 1 but to  %2.9g \n", (double)sum);
		}}
	fprintf(cFpWriteModel, "\n\tT: ");
	n=0;sum=0;
	for (t=0; t<nbStates; t++) 
		{
		if (trProb[s][t]>0)
			{
			n++;
			sum += trProb[s][t];
			fprintf(cFpWriteModel, "  %s %g  ",  statenames[t], (double)trProb[s][t]);
			}
		}
	if (n==0)  fprintf(cFpWriteModel, "  END 0");
	if ((sum > ktol2 ) || (sum < ktol1 ))
		{
		if (strncmp(statenames[s],"END",3) != 0) 
			{
			fprintf(stderr, "\nWarning: transition probabilities of state %s  ", statenames[s]);
			fprintf(stderr, "do not sum up to 1 but to  %2.9g \n", (double)sum);
		}}
	}
fprintf(cFpWriteModel, "\n################################################");
fprintf(cFpWriteModel, "\n#\nNULLMODEL: ");
for (t=1; t<=nbLetters ; t++)	{fprintf(cFpWriteModel, " %g ",nullEmFreq[t] );}
fprintf(cFpWriteModel, "\n#\nfix_tr: ");
for (t=0; t<nbStates; t++) 
	{ 
	if (updatetr[t]== 0)
		{
		fprintf(cFpWriteModel, " %s ",  statenames[t]);
	}	}

fprintf(cFpWriteModel, "\n#\nfix_em: ");
for (t=0; t<nbStates; t++) 
	{ 
	if (updateem[t]== 0)
		{
		fprintf(cFpWriteModel, " %s ",  statenames[t]);
	}	}

for (t=0; t<nbStates; t++) 
	{ 
	if (weightEM[t]!= 1)
		{
		fprintf(cFpWriteModel, "\nweight_em: ");
		fprintf(cFpWriteModel, " %g", (double)weightEM[t]);
		fprintf(cFpWriteModel, " %s ",  statenames[t]);
	}	}


fprintf(cFpWriteModel, "\n#");
for (t=1; t<=nbEMties; t++) 
	{ 
	fprintf(cFpWriteModel, "\ntie_em: ");
	for (s=1; s<=ties[t][0]; s++) 
		{
		fprintf(cFpWriteModel, " %s ",  statenames[ties[t][s]]);
	}	}
for (t=1; t<=nbcEMties; t++) 
	{ 
	fprintf(cFpWriteModel, "\ncompl_em: ");
	fprintf(cFpWriteModel, " %s ",  statenames[cties[t][0]]);
	fprintf(cFpWriteModel, " %s ",  statenames[cties[t][1]]);
	}	



fprintf(cFpWriteModel, "\n#\n#avoid states being in more than one tie group!\n#");
fprintf(cFpWriteModel, "\n#avoid blank lines, especially a trailing blank line\n#######");
if (verbose) { cerr << "\n ending ModelFB::WriteModelFB\n"; }
}

/* WriteModelFB
************************************/

// ---------------------------------------------------------------------------
//	 RndEmission
// ---------------------------------------------------------------------------
int  ModelFB::RndEmission(const int s)  
{
double cumem[kMaxNbLetters+1];
double r, z;
int t, last;
cumem[0]=0;last=0;
for (t=1; t<=nbLetters; t++)
	{
	cumem[t]=cumem[t-1]+ emProb[s][t];//COULD DO ONLY ONCE AND STORE PER EACH STATE
	if (emProb[s][t] > 0) last=t;
	}
for (t=1; t<=nbLetters; t++)
	{
	cumem[t]=cumem[t] / cumem[last];//COULD DO ONLY ONCE AND STORE PER EACH STATE
	}
if 	(cumem[last] != 1) 
	 {
	 if (verbose) {cerr << "Warning cumem[last] not = 1! in ModelFB::RndEmission for state " << s << endl;}
	 cumem[last] = 1;
	 }
z = 0;
while (z<= 0) 
	{
	r=((double) ran2(seedran2));
	t=1;
	while ((r>cumem[t]) && (t<nbLetters)) t++;
	z= cumem[t]; 
	}
return (t);
}
/* RndEmission
************************************/

// ---------------------------------------------------------------------------
//	 RndStateTransition
// ---------------------------------------------------------------------------
int  ModelFB::RndStateTransition(const int s)  
{
double cumtr[kMaxNbStates+1];
double r, z;
int t, last;
cumtr[0]=trProb[s][0];last=0;
for (t=1; t<nbStates; t++)
	{
	cumtr[t]=cumtr[t-1]+ trProb[s][t];  
	if (trProb[s][t] > 0) last=t;
	}
	
for (t=1; t<nbStates; t++)
	{
	cumtr[t]=cumtr[t] /cumtr[last]; 
	}
if 	(cumtr[last]!= 1)  
	{
	if (verbose)  {cerr << "Warning cumtr[last]  not = 1! in ModelFB::RndStateTransition for state " << s << endl;}
	cumtr[last] = 1;
	}
z = 0;
while (z<= 0) 
	{
	t=0;
	r=((double) ran2(seedran2));
	while ((r>cumtr[t]) && (t<(nbStates-1))) t++;
	z= cumtr[t];
	}
return (t);
}
/* RndStateTransition
************************************/

// ---------------------------------------------------------------------------
//	 GenerateSequence
// ---------------------------------------------------------------------------
void  ModelFB::GenerateSequence (FILE *cFpWriteGenSeq, int genseqnb, int maxlen)
{
  int pos,state,l;

  if (verbose) cerr <<"\n ModelFB::GenerateSequence" << endl;

  pos=0;
  state=startstate;
  if (bfileoutput)
  {
  if (genseqnb==1) 
	  {
	  fprintf(cFpWriteGenSeq,"sequence\tnposition\tstateNB\tstateNAME\tletter");
	  }
  fprintf(cFpWriteGenSeq,"\n%d\t%d\t%d\t%s\tNA",genseqnb, pos, state, statenames[state]);
  }
  state=RndStateTransition(state);

  while (state != endstate)
    {
     pos++;

    if (pos>1 && (pos-1)%kLineLenRndSeqs==0)
	cout << "\n";

    if (maxlen>0 && pos>maxlen) {  
	cout << "*";
	break;
      }
      l=RndEmission(state);
      cout << letter[l];
      if (bfileoutput) fprintf(cFpWriteGenSeq,"\n%d\t%d\t%d\t%s\t%c",genseqnb,pos,state, statenames[state],letter[l]);
      state=RndStateTransition(state);
    }

  if (bfileoutput)
  	{
	  if (state==endstate)
  		{fprintf(cFpWriteGenSeq,"\n%d\t%d\t%d\t%s\tNA",genseqnb,pos, state, statenames[state]);}
	  else
  		{fprintf(cFpWriteGenSeq,"\n%d\t%d\tNA\tTRUNCATED\tNA",genseqnb,pos);}
    }
}

// ---------------------------------------------------------------------------
//	 GenerateSequence
// ---------------------------------------------------------------------------
void  ModelFB::GenerateSequenceB (int genseqnb, int maxlen)
{
  int pos,state,l;

  if (verbose) cerr <<"\n ModelFB::GenerateSequenceB" << endl;
  pos=0;
  state=startstate;
  state=RndStateTransition(state);
  while (state != endstate)
    {
     pos++;
    if (pos>1 && (pos-1)%kLineLenRndSeqs==0)
	cout << "\n";
    if (maxlen>0 && pos>maxlen) {  
	cout << "*";
	break;
      }
      l=RndEmission(state);
      cout << letter[l];
      state=RndStateTransition(state);
    }
}
/* GenerateSequence
************************************/

// ---------------------------------------------------------------------------
//	 costructor ModelFB	//constructor, initialise static memory
// ---------------------------------------------------------------------------
ModelFB::ModelFB ()
	{;
	}

/* costructor ModelFB
************************************/

// ---------------------------------------------------------------------------
//	 costructor ModelFB	//constructor, initialise dynamic memory
// ---------------------------------------------------------------------------
void ModelFB::AssignMemory (int modus, int s, int a, int l, int m, int t, int T, int c)
	{
	int i,j;
	nbLetters=a;
	nbStates=s;
	maxSeqLength= l;
	MaxNameLen=m;
	int nbTies=t;
	int maxnbTiesperGroup=T;
	int nbcTies=c;
	// modus: 0 : FBmain  1 : GenMain
	// 2 : if ((bdsDNA) || (bdsDNAjointly)) BWmaindsDNA; 
	// 3 : FBmain	  
	//4 : Vitmain
	//5 :  VitLearnMain
	
	letter[0]='X';

if (verbose) cerr << "start Ass Mem" << endl; 

	ctiesmem=(int *) malloc( (nbcTies+1)*2*sizeof(int));
	if (ctiesmem==NULL) {cerr<<"\nFailed in malloc ctiesmem"<<endl;exit(0);}
	cties=(int **) malloc((nbcTies+1)*sizeof(int *));
	if (cties==NULL) {cerr<<"\nFailed in malloc cties"<<endl;exit(0);}
	for (i=0; i<=nbcTies; i++) cties[i]=ctiesmem+i*2;

	tiesmem=(int *) malloc( (nbTies+1)*(maxnbTiesperGroup+1)*sizeof(int));
	if (tiesmem==NULL) {cerr<<"\nFailed in malloc tiesmem"<<endl;exit(0);}
	ties=(int **) malloc((nbTies+1)*sizeof(int *));
	if (ties==NULL) {cerr<<"\nFailed in malloc ties"<<endl;exit(0);}
	for (i=0; i<=nbTies; i++) ties[i]=tiesmem+i*(maxnbTiesperGroup+1);
	for (i=0; i<=nbTies; i++)
		{
		ties[i][0]=0;// nb of ties in that tie group
		}

	trProbmem=(long double *) malloc(nbStates*nbStates*sizeof(long double));
	if (trProbmem==NULL) {cerr<<"\nFailed in malloc trProbmem"<<endl;exit(0);}
	trProb=(long double **) malloc(nbStates*sizeof(long double *));
	if (trProb==NULL) {cerr<<"\nFailed in malloc trProb"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) trProb[i]=trProbmem+i*nbStates;
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<nbStates; j++) trProb[i][j]=0;
		}

	emProbmem=(long double *) malloc(nbStates*(nbLetters+1)*sizeof(long double));
	if (emProbmem==NULL) {cerr<<"\nFailed in malloc emProbmem"<<endl;exit(0);}
	emProb=(long double **) malloc(nbStates*sizeof(long double *));
	if (emProb==NULL) {cerr<<"\nFailed in malloc emProb"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) emProb[i]=emProbmem+i*(nbLetters+1);
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<=nbLetters; j++)  emProb[i][j]=0;
		}

	Contmem=(long double *) malloc(nbStates*(nbLetters+1)*sizeof(long double));
	if (Contmem==NULL) {cerr<<"\nFailed in malloc Contmem"<<endl;exit(0);}
	Cont=(long double **) malloc(nbStates*sizeof(long double *));
	if (Cont==NULL) {cerr<<"\nFailed in malloc Cont"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) Cont[i]=Contmem+i*(nbLetters+1);
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<=nbLetters; j++)  Cont[i][j]=0;
		}

	tContmem=(long double *) malloc(nbStates*nbStates*sizeof(long double));
	if (tContmem==NULL) {cerr<<"\nFailed in malloc tContmem"<<endl;exit(0);}
	tCont=(long double **) malloc(nbStates*sizeof(long double *));
	if (tCont==NULL) {cerr<<"\nFailed in malloc tCont"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) tCont[i]=tContmem+i*nbStates;
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<nbStates; j++) tCont[i][j]=0;
		}

if ((bdsDNA) || (bdsDNAjointly))
	{
	Contmem2=(long double *) malloc(nbStates*(nbLetters+1)*sizeof(long double));
	if (Contmem2==NULL) {cerr<<"\nFailed in malloc Contmem2"<<endl;exit(0);}
	Cont2=(long double **) malloc(nbStates*sizeof(long double *));
	if (Cont2==NULL) {cerr<<"\nFailed in malloc Cont2"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) Cont2[i]=Contmem2+i*(nbLetters+1);
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<=nbLetters; j++)  Cont2[i][j]=0;
		}

	tContmem2=(long double *) malloc(nbStates*nbStates*sizeof(long double));
	if (tContmem2==NULL) {cerr<<"\nFailed in malloc tContmem2"<<endl;exit(0);}
	tCont2=(long double **) malloc(nbStates*sizeof(long double *));
	if (tCont2==NULL) {cerr<<"\nFailed in malloc tCont2"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) tCont2[i]=tContmem2+i*nbStates;
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<nbStates; j++) tCont2[i][j]=0;
		}
	} //if ((bdsDNA) || (bdsDNAjointly))


if (modus>3) 
{
	tracebackMmem=(int*) malloc(nbStates*(kMaxVitSeqLen)*sizeof(int));
	if (tracebackMmem==NULL) {cerr<<"\nFailed in malloc tracebackMmem"<<endl;exit(0);}
	tracebackM =(int **) malloc(nbStates*sizeof(int*));
	if (tracebackM==NULL) {cerr<<"\nFailed in malloc tracebackM"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) tracebackM[i]=tracebackMmem+i*(kMaxVitSeqLen);
}


if ((modus<=3) && (modus!=1) )
{
	scForwmem=(long double *) malloc(nbStates*(maxSeqLength+2)*sizeof(long double));
	if (scForwmem==NULL) {cerr<<"\nFailed in malloc scForwmem"<<endl;exit(0);}
	scForw=(long double **) malloc(nbStates*sizeof(long double *));
	if (scForw==NULL) {cerr<<"\nFailed in malloc scForw"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) scForw[i]=scForwmem+i*(maxSeqLength+2);
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<=maxSeqLength+1; j++)  scForw[i][j]=0;
		}

	scBackwmem=(long double *) malloc(nbStates*(maxSeqLength+2)*sizeof(long double));
	if (scBackwmem==NULL) {cerr<<"\nFailed in malloc scBackwmem"<<endl;exit(0);}
	scBackw=(long double **) malloc(nbStates*sizeof(long double *));
	if (scBackw==NULL) {cerr<<"\nFailed in malloc scBackw"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) scBackw[i]=scBackwmem+i*(maxSeqLength+2);
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<=maxSeqLength+1; j++)  scBackw[i][j]=0;
		}

	PosStateProbmem=(double *) malloc((maxSeqLength+3)*nbStates*sizeof(double));
	if (PosStateProbmem==NULL) {cerr<<"\nFailed in malloc PosStateProbmem"<<endl; exit(0);}
	PosStateProb=(double **) malloc((maxSeqLength+3)*sizeof(double *));
	if (PosStateProb==NULL) {cerr<<"\nFailed in malloc PosStateProb"<<endl; exit(0);}
	for (i=0; i<(maxSeqLength+3); i++) PosStateProb[i]=PosStateProbmem+i*nbStates;
	for (i=0; i<(maxSeqLength+3); i++)
		{
		for (j=0; j<nbStates; j++)  PosStateProb[i][j]=0;
		}
}


if ((modus==2) || (modus==5))
{
	scSumEmTotmem=(long double *) malloc(nbStates*(nbLetters+1)*sizeof(long double));
	if (scSumEmTotmem==NULL) {cerr<<"\nFailed in malloc scSumEmTotmem"<<endl;exit(0);}
	scSumEmTot=(long double **) malloc(nbStates*sizeof(long double *));
	if (scSumEmTot==NULL) {cerr<<"\nFailed in malloc emProb"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) scSumEmTot[i]=scSumEmTotmem+i*(nbLetters+1);
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<=nbLetters; j++)  scSumEmTot[i][j]=0;
		}

	scSumTrTotmem=(long double *) malloc(nbStates*nbStates*sizeof(long double));
	if (scSumTrTotmem==NULL) {cerr<<"\nFailed in malloc scSumTrTot"<<endl;exit(0);}
	scSumTrTot=(long double **) malloc(nbStates*sizeof(long double *));
	if (scSumTrTot==NULL) {cerr<<"\nFailed in malloc trProb"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) scSumTrTot[i]=scSumTrTotmem+i*nbStates;
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<nbStates; j++) scSumTrTot[i][j]=0;
		}
}

if (bUseExtP)
{
	extPmem=(double *) malloc(nbStates*(kMaxExtPLen)*sizeof(double));
	if (extPmem==NULL) {cerr<<"\nFailed in malloc extPmem"<<endl;exit(0);}
	extP=(double **) malloc(nbStates*sizeof(double *));
	if (extP==NULL) {cerr<<"\nFailed in malloc extP"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) extP[i]=extPmem+i*(kMaxExtPLen);
}

	scEmProbmem=(long double *) malloc(nbStates*(nbLetters+1)*sizeof(long double));
	if (scEmProbmem==NULL) {cerr<<"\nFailed in malloc scEmProbmem"<<endl;exit(0);}
	scEmProb=(long double **) malloc(nbStates*sizeof(long double *));
	if (scEmProb==NULL) {cerr<<"\nFailed in malloc scEmProb"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) scEmProb[i]=scEmProbmem+i*(nbLetters+1);
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<=nbLetters; j++)  scEmProb[i][j]=0;
		}

	scTrProbmem=(long double *) malloc(nbStates*nbStates*sizeof(long double));
	if (scTrProbmem==NULL) {cerr<<"\nFailed in malloc scTrProbmem"<<endl;exit(0);}
	scTrProb=(long double **) malloc(nbStates*sizeof(long double *));
	if (scTrProb==NULL) {cerr<<"\nFailed in malloc scTrProb"<<endl;exit(0);}
	for (i=0; i<nbStates; i++) scTrProb[i]=scTrProbmem+i*nbStates;
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<nbStates; j++) scTrProb[i][j]=0;
		}

	statenamesmem=(char *) malloc(nbStates*MaxNameLen*sizeof(char));
	if (statenamesmem==NULL) {cerr<<"\nFailed in malloc statenamesmem"<<endl; exit(0);}
	statenames=(char **) malloc(nbStates*sizeof(char *));
	if (statenames==NULL) {cerr<<"\nFailed in malloc statenames"<<endl; exit(0);}
	for (i=0; i<nbStates; i++) statenames[i]=statenamesmem+i*MaxNameLen;
	for (i=0; i<nbStates; i++)
		{
		for (j=0; j<MaxNameLen; j++)  statenames[i][j]='\0';
		}

//single dimensional arrays
	pseudocountsEm=(double *) malloc((nbLetters+1)*sizeof(double));
	nullEmFreq=(double *) malloc((nbLetters+1)*sizeof(double));
	weightEM=(double *) malloc((nbStates)*sizeof(double));
	for (j=0; j<nbStates; j++) weightEM[j]=1;

	ForwConnList=(int *) malloc(2*nbStates*nbStates*sizeof(int));
	BackConnList=(int *) malloc(2*nbStates*nbStates*sizeof(int));
	ForwConnFirst=(int *) malloc(2*nbStates*sizeof(int));
	ForwConnLast=(int *) malloc(2*nbStates*sizeof(int));
	BackConnFirst=(int *) malloc(2*nbStates*sizeof(int));
	BackConnLast=(int *) malloc(2*nbStates*sizeof(int));
	updatetr=(int *) malloc(2*nbStates*sizeof(int));	
	pseudow=(int *) malloc(2*nbStates*sizeof(int));
	updateem=(int *) malloc(2*nbStates*sizeof(int));
	for (j=0; j<nbStates; j++) 
		{
		 updatetr[j]=1;updateem[j]=1;pseudow[j]=1;
		 }

if (verbose) cerr << "end Ass Mem" << endl; 
	}

/* constructor ModelFB
************************************/

// ---------------------------------------------------------------------------
//	 destructor ~ModelFB 		/destructor; free dynamically assigned memory
// ---------------------------------------------------------------------------
ModelFB::~ModelFB ()
{}
//  this created some segmentation faults
	
/*	free(cties);free(ctiesmem);
	free(ties);free(tiesmem);
	free(emProb);free(emProbmem);
	free(trProb);free(trProbmem);
	free(Cont);free(Contmem);
	free(tCont);free(tContmem);
	free(statenames);free(statenamesmem);
	free(scEmProb);free(scEmProbmem);
	free(scTrProb);free(scTrProbmem);

	free(nullEmFreq);free(pseudocountsEm);
	free(BackConnList);free(ForwConnList);
	free(updatetr);free(updateem);
	free(BackConnLast);free(BackConnFirst);
	free(ForwConnLast);free(ForwConnFirst);
	free(pseudocountsEm);

if ((usemodus==2) || (usemodus==5))
	{	
	free(scSumEmTot);free(scSumEmTotmem);
	free(scSumTrTot);free(scSumTrTotmem);
	}

if ((usemodus<=3) && (usemodus!=1))
	{
	free(scForw);free(scForwmem);
	free(scBackw);free(scBackwmem);
	free(PosStateProb);free(PosStateProbmem);
	}
	
if (usemodus>3)
	{ free(tracebackM);free(tracebackMmem);}

if ((bdsDNA) || (bdsDNAjointly))
	{
	free(Cont2);free(Contmem2);
	free(tCont2);free(tContmem2);
	}
}
	
// destructor ~ModelFB
************************************/

// ---------------------------------------------------------------------------
//	 SetUpConnections 		
// ---------------------------------------------------------------------------
void  ModelFB::SetUpConnections(void)
{
int   t, s, n;
bool first, checkconnections;
checkconnections = true;
if (CutoffForConnecting > 0)  checkconnections = true;
if (verbose)  cerr << "\n start SetUpConnections " << endl;

n=-1;
for (s=0; s < nbStates; s++)
	{
	ForwConnFirst[s] = -1; ForwConnLast[s] = -2;
	first=true;
	for (t=0; t < nbStates ; t++)
		{
		if (trProb[s][t] > kMinProb)
			{
			ForwConnList[++n] = t;
			ForwConnLast[s] = n;
			if (first==true) {first=false; ForwConnFirst[s] = n;}
	}	}	}

n=-1;
for (s=0; s < nbStates; s++)
	{
	BackConnFirst[s] = -1; BackConnLast[s] = -2;
	first=true;
	for (t=0; t < nbStates ; t++)
		{
		if (trProb[t][s] > kMinProb)
			{
			BackConnList[++n] = t;
			BackConnLast[s] = n;
			if (first==true) {first=false; BackConnFirst[s]= n;}
	}	}	}
	
if (checkconnections)
	{
	for (s=0; s < nbStates; s++)
		{
		for (n=ForwConnFirst[s]; n<= ForwConnLast[s]; n++)
			{
			t = ForwConnList[n];
	}	}	

	for (s=0; s < nbStates; s++)
		{
		for (n=BackConnFirst[s]; n<=BackConnLast[s] ; n++)
			{
			t = BackConnList[n];
	}	}	}	
if (verbose) cerr << "end SetUpConnections " << endl;
}
/* SetUpConnections
************************************/

// ---------------------------------------------------------------------------
//	 SetUpScaledParameters 			
// ---------------------------------------------------------------------------
void ModelFB::SetUpScaledParameters(void)
{
int	i,j; 
long double a, t, b, r;
if (verbose) cerr << "\nstart SetUpScaledParameters " << endl;

a=0;b=0;
for (i=0; i<nbStates; i++)
	{
	r=0;
	for (j=0; j<=nbLetters; j++)
		{
		r += emProb[i][j] * emProb[i][j];
		} 
	b+=r;
	
	t=0;	
	for (j=0; j<nbStates; j++)
		{
		t += trProb[i][j] * trProb[i][j];
		}
	a+=t;
	}
a=(a+0.001)/(nbStates-1);
b=b/(nbStates-2);
emScale = (1/a) * (1/b) * 1.2;
if (verbose) printf("\n emScale = %g ", (double)emScale);

for (i=0; i<nbStates; i++)
	{
	for (j=0; j<=nbLetters; j++)
		{
		scEmProb[i][j] =  emScale * emProb[i][j];// scale all em Probs uniformly
		} 
		
	for (j=0; j<nbStates; j++)
		{
		scTrProb[i][j] = trProb[i][j];// no scaling, added factor to emscale
		}
	}
	
if (verbose) cerr << "\nend SetUpScaledParameters " << endl;
}
/* SetUpWeightedParameters
************************************/

// ---------------------------------------------------------------------------
//	 SetUpScaledParametersW SetUpWeightedScaledParameters 			
// ---------------------------------------------------------------------------
void ModelFB::SetUpScaledParametersW(void)
{
int	i,j; 	
long double a, t, b, r;
if (verbose) cerr << "\nstart SetUpScaledParametersW: SetUpWeightedScaledParameters " << endl;

for (i=0; i<nbStates; i++)
	{
	for (j=0; j<=nbLetters; j++)
		{
		emProb[i][j] =  pow ( emProb[i][j], weightEM[i]);
		} 
	}

a=0;b=0;
for (i=0; i<nbStates; i++)
	{
	r=0;
	for (j=0; j<=nbLetters; j++)
		{
		r += emProb[i][j] * emProb[i][j];
		} 
	b+=r;
	
	t=0;	
	for (j=0; j<nbStates; j++)
		{
		t += trProb[i][j] * trProb[i][j];
		}
	a+=t;
	}
a=(a+0.001)/(nbStates-1);
b=b/(nbStates-2);
emScale = (1/a) * (1/b) * 1.2;
if (verbose) printf("\n emScale = %g ", (double)emScale);

for (i=0; i<nbStates; i++)
	{
	for (j=0; j<=nbLetters; j++)
		{
		scEmProb[i][j] =  emScale * emProb[i][j];// scale all em Probs uniformly
		} 
		
	for (j=0; j<nbStates; j++)
		{
		scTrProb[i][j] = trProb[i][j];// no scaling, added factor to emscale
		}
	}
	

if (verbose) cerr << "\nend SetUpScaledParametersW " << endl;
}

/* SetUpWeightedParameters
************************************/

// ---------------------------------------------------------------------------
//	 SetUpLogParametersW: SetUp Weighted LogParameters 			
// ---------------------------------------------------------------------------
void ModelFB::SetUpLogParametersW(void)
{
int	i,j; 	
if (verbose) cerr << "\nstart SetUpLogParametersW: SetUp Weighted LogParameters" << endl;

for (i=0; i<nbStates; i++)
	{
	for (j=0; j<=nbLetters; j++)
		{
		scEmProb[i][j] = weightEM[i] * scEmProb[i][j];
		} 
	}
	
if (verbose) cerr << "\nend SetUpLogParametersW" << endl;
}
/* SetUpScaledParameters
************************************/

// ---------------------------------------------------------------------------
//	 model.SetUpLogParameters
// ---------------------------------------------------------------------------
void ModelFB::SetUpLogParameters(void)
{
int	i,j; 
if (verbose) cerr << "\n start SetUpLogParameters " << endl;

for (i=0; i<nbStates; i++)
	{
	for (j=0; j<=nbLetters; j++)
		{
		if (emProb[i][j]>0) 
			{
			scEmProb[i][j] =  log(emProb[i][j]);
			}
		else  scEmProb[i][j] =  -100000;
		} 
		
	for (j=0; j<nbStates; j++)
		{
		if (trProb[i][j]>0) 
			{
			scTrProb[i][j] = log(trProb[i][j]);
			}
		else  scTrProb[i][j] =  -100000;
		}
	}
}

/*	a scale up to before logging can be  used to keep the sum of logs near zero */
/* SetUpLogParameters
************************************/

// ---------------------------------------------------------------------------
//	  Viterbi algorithm
// --------------------------------------------------------- ------------------

void ModelFB::Viterbi (int seqNb, int totalNumber, const int Seq[], FILE *cFpWriteVITDEC) 
{
int			i, s, ps, n, k, traceback=0; 
long double  	max=0, best[nbStates], newBest[nbStates], alt; 
bool	accessed[nbStates], newAccessed[nbStates];

if (verbose) cerr << "\n start Viterbi" << endl;

for (s=0; s<nbStates; s++) {accessed[s] = false; tracebackM[s][0]=-1; best[s]=5;}
best[startstate] = 0; 
accessed[startstate] = true;possible=true;
for (i=1; i<=totalNumber; i++)  
	{ 
	if (possible)
		{
		possible=false;
		for (s=0; s<nbStates; s++) 
			{
			k=0;newAccessed[s] =false;
			for (n=BackConnFirst[s]; n<=BackConnLast[s]; n++)
				{
				ps = BackConnList[n]; 
				if (accessed[ps]) 
					{
					alt = best[ps] + scTrProb[ps][s];
					if (k==0) {max = alt; traceback = ps;  k++;}
					else {if (alt > max)	{max = alt;  traceback = ps;}}
					}
				}			
			if ((k>0) && (emProb[s][Seq[i]]>0)) 
				{
				possible=true;newAccessed[s] =true;
				newBest[s]= max + scEmProb[s][Seq[i]]; 
				tracebackM[s][i] = traceback;
				}
			else {newAccessed[s] = false;tracebackM[s][i]=-1;  newBest[s]=5;}
			}
		for (s=0; s<nbStates; s++) {best[s] = newBest[s];  accessed[s] = newAccessed[s] ;}
	}	}

if (possible)
	{
	k=0;possible=false;
	s = endstate; i=totalNumber+1;
	for (n=BackConnFirst[s]; n<=BackConnLast[s]; n++)
		{
		ps = BackConnList[n]; 
		if (accessed[ps]) 
			{
			alt = best[ps] + scTrProb[ps][s];
			if (k==0) {max = alt; traceback = ps;  k++; possible=true;}
			else {if (alt > max)	{max = alt;  traceback = ps;}}
			}
		}
	}
if (possible)
	{	
	tracebackM[s][i] = traceback;
	ViterbiPathLogProb = max;
	 //RECONSTRUCT THE PATH
	bestPathM[totalNumber+1] = endstate;
	i=totalNumber+1;
	bestPathM[i-1] = tracebackM[bestPathM[i]][i];
	for (i=totalNumber; i>=1; i--) 	
		{
		bestPathM[i-1] = tracebackM[bestPathM[i]][i];   
		}

		if (seqNb==1) {fprintf(cFpWriteVITDEC,"Nsequence\tNposition\tSEQ\tNstate\tstate_name");}
		for (i=0; i<=totalNumber+1; i++) 	
			{
		fprintf(cFpWriteVITDEC,"\n%d\t%d\t%c\t%d\t%s", seqNb,i,Decode[Seq[i]],bestPathM[i], statenames[bestPathM[i]]);
			}
	}
else
	{
	ViterbiPathLogProb = -5000000000.;  //  fictitious, for vit learning protocol
	fprintf(cFpWriteVITDEC, "\nlog viterbi-probability = -INF  \nWarning: Sequence %d has Probability 0;", seqNb);
	cerr <<  "\nWarning: Sequence Nb  " << seqNb  << "  has Probability 0"  << endl;
	i=totalNumber+1; tracebackM[s][i] = startstate;
	for (i=totalNumber; i>=1; i--) 	{bestPathM[i-1] = startstate;  }
	}
//bestPathM[0] = startstate; 
}
/* Viterbi 
************************************/


// ---------------------------------------------------------------------------
//	  Viterbi algorithmViterbiExtP
// --------------------------------------------------------- ------------------

void ModelFB::ViterbiExtP (int seqNb, int totalNumber, const int Seq[], FILE *cFpWriteVITDEC) 
{
int			i, s, ps, n, k, traceback=0; 
long double  	max=0, best[nbStates], newBest[nbStates], alt; 
bool	accessed[nbStates], newAccessed[nbStates];

if (verbose) cerr << "\n start ViterbiExtP" << endl;
for (s=0; s<nbStates; s++) {accessed[s] = false; tracebackM[s][0]=-1; best[s]=5;}
best[startstate] = 0;
accessed[startstate] = true;possible=true;
for (i=1; i<=totalNumber; i++)  
	{ 
	if (possible)
		{
		possible=false;
		for (s=0; s<nbStates; s++) 
			{
			k=0;newAccessed[s] =false;
			for (n=BackConnFirst[s]; n<=BackConnLast[s]; n++)
				{
				ps = BackConnList[n]; 
				if (accessed[ps]) 
					{
					alt = best[ps] + scTrProb[ps][s];
					if (k==0) {max = alt; traceback = ps;  k++;}
					else {if (alt > max)	{max = alt;  traceback = ps;}}
					}
				}			
			if ((k>0) && (emProb[s][Seq[i]]>0) && (extP[s][i]>0) ) 
				{
				possible=true;newAccessed[s] =true;
				newBest[s]= max + scEmProb[s][Seq[i]] + log(extP[s][i]);
				tracebackM[s][i] = traceback;
				}
			else {newAccessed[s] = false;tracebackM[s][i]=-1;  newBest[s]=5;}
			}
		for (s=0; s<nbStates; s++) {best[s] = newBest[s];  accessed[s] = newAccessed[s] ;}
	}	}
if (possible)
	{
	k=0;possible=false;
	s = endstate; i=totalNumber+1;
	for (n=BackConnFirst[s]; n<=BackConnLast[s]; n++)
		{
		ps = BackConnList[n]; 
		if (accessed[ps]) 
			{
			alt = best[ps] + scTrProb[ps][s];
			if (k==0) {max = alt; traceback = ps;  k++; possible=true;}
			else {if (alt > max)	{max = alt;  traceback = ps;}}
			}
		}
	}
if (possible)
	{	
	tracebackM[s][i] = traceback;
	ViterbiPathLogProb = max;
	 //RECONSTRUCT THE PATH
	bestPathM[totalNumber+1] = endstate;
	i=totalNumber+1;
	bestPathM[i-1] = tracebackM[bestPathM[i]][i];
	for (i=totalNumber; i>=1; i--) 	
		{
		bestPathM[i-1] = tracebackM[bestPathM[i]][i];   
		}

		if (seqNb==1) {fprintf(cFpWriteVITDEC,"Nsequence\tNposition\tSEQ\tNstate\tstate_name");}
		for (i=0; i<=totalNumber+1; i++) 	
			{
		fprintf(cFpWriteVITDEC,"\n%d\t%d\t%c\t%d\t%s", seqNb,i,Decode[Seq[i]],bestPathM[i], statenames[bestPathM[i]]);
			}
	}
else
	{
	ViterbiPathLogProb = -5000000000.;  //  fictitious, for vit learning protocol
	fprintf(cFpWriteVITDEC, "\nlog viterbi-probability = -INF  \nWarning: Sequence %d has Probability 0;", seqNb);
	cerr <<  "\nWarning: Sequence Nb  " << seqNb  << "  has Probability 0"  << endl;
	i=totalNumber+1; tracebackM[s][i] = startstate;
	for (i=totalNumber; i>=1; i--) 	{bestPathM[i-1] = startstate;  }
	}
}
/* ViterbiExtP 
************************************/

// ---------------------------------------------------------------------------
//	  Viterbi algorithm for learning 
// --------------------------------------------------------- ------------------

void ModelFB::ViterbiL (int seqNb, int totalNumber, const int Seq[]) 
{
int			i, s, ps, n, k, traceback=0; 
long double  	max=0, best[nbStates], newBest[nbStates], alt; 
bool	accessed[nbStates], newAccessed[nbStates];

for (s=0; s<nbStates; s++) {accessed[s] = false; tracebackM[s][0]=-1; best[s]=5;}
best[startstate] = 0;
accessed[startstate] = true;possible=true;
for (i=1; i<=totalNumber; i++)  
	{ 
	if (possible)
		{
		possible=false;
		for (s=0; s<nbStates; s++) 
			{
			k=0;newAccessed[s] =false;
			for (n=BackConnFirst[s]; n<=BackConnLast[s]; n++)
				{
				ps = BackConnList[n]; 
				if (accessed[ps]) 
					{
					alt = best[ps] + scTrProb[ps][s];
					if (k==0) {max = alt; traceback = ps;  k++;}
					else {if (alt > max)	{max = alt;  traceback = ps;}}
					}
				}			
			if ((k>0) && (emProb[s][Seq[i]]>0)) 
				{
				possible=true;newAccessed[s] =true;
				newBest[s]= max + scEmProb[s][Seq[i]]; 
				tracebackM[s][i] = traceback;
				}
			else {newAccessed[s] = false;tracebackM[s][i]=-1;  newBest[s]=5;}
			}
		for (s=0; s<nbStates; s++) {best[s] = newBest[s];  accessed[s] = newAccessed[s] ;}
	}	}
if (possible)
	{
	k=0;possible=false;
	s = endstate; i=totalNumber+1;
	for (n=BackConnFirst[s]; n<=BackConnLast[s]; n++)
		{
		ps = BackConnList[n]; 
		if (accessed[ps]) 
			{
			alt = best[ps] + scTrProb[ps][s];
			if (k==0) {max = alt; traceback = ps;  k++; possible=true;}
			else {if (alt > max)	{max = alt;  traceback = ps;}}
			}
		}
	}
if (possible)
	{	
	tracebackM[s][i] = traceback;
	ViterbiPathLogProb = max;
	 //RECONSTRUCT THE PATH	 					
	bestPathM[totalNumber+1] = endstate;
	i=totalNumber+1;
	bestPathM[i-1] = tracebackM[bestPathM[i]][i];
	for (i=totalNumber; i>=1; i--) 	
		{
		bestPathM[i-1] = tracebackM[bestPathM[i]][i];   
		}
	}
else
	{
	ViterbiPathLogProb = -5000000000.;  //  fictitious, for vit learning protocol
	cerr <<  "\nWarning: Sequence Nb  " << seqNb  << "  has Probability 0"  << endl;
	i=totalNumber+1; tracebackM[s][i] = startstate;
	for (i=totalNumber; i>=1; i--) 	{bestPathM[i-1] = startstate;  }
	}
//bestPathM[0] = startstate; 
}
// ---------------------------------------------------------------------------



// ---------------------------------------------------------------------------
//	  Viterbi algorithm for learning 
// --------------------------------------------------------- ------------------
void ModelFB::ViterbiLextP (int seqNb, int totalNumber, const int Seq[]) 
{
int			i, s, ps, n, k, traceback=0; 
long double  	max=0, best[nbStates], newBest[nbStates], alt; 
bool	accessed[nbStates], newAccessed[nbStates];

for (s=0; s<nbStates; s++) {accessed[s] = false; tracebackM[s][0]=-1; best[s]=5;}
best[startstate] = 0;
accessed[startstate] = true;possible=true;
for (i=1; i<=totalNumber; i++)  
	{ 
	if (possible)
		{
		possible=false;
		for (s=0; s<nbStates; s++) 
			{
			k=0;newAccessed[s] =false;
			for (n=BackConnFirst[s]; n<=BackConnLast[s]; n++)
				{
				ps = BackConnList[n]; 
				if (accessed[ps]) 
					{
					alt = best[ps] + scTrProb[ps][s];
					if (k==0) {max = alt; traceback = ps;  k++;}
					else {if (alt > max)	{max = alt;  traceback = ps;}}
					}
				}			
			if ((k>0) && (emProb[s][Seq[i]]>0) && (extP[s][i]>0) ) 
				{
				possible=true;newAccessed[s] =true;
				newBest[s]= max + scEmProb[s][Seq[i]] + log(extP[s][i]) ;
				tracebackM[s][i] = traceback;
				}
			else {newAccessed[s] = false;tracebackM[s][i]=-1;  newBest[s]=5;}
			}
		for (s=0; s<nbStates; s++) {best[s] = newBest[s];  accessed[s] = newAccessed[s] ;}
	}	}
if (possible)
	{
	k=0;possible=false;
	s = endstate; i=totalNumber+1;
	for (n=BackConnFirst[s]; n<=BackConnLast[s]; n++)
		{
		ps = BackConnList[n]; 
		if (accessed[ps]) 
			{
			alt = best[ps] + scTrProb[ps][s];
			if (k==0) {max = alt; traceback = ps;  k++; possible=true;}
			else {if (alt > max)	{max = alt;  traceback = ps;}}
			}
		}
	}
if (possible)
	{	
	tracebackM[s][i] = traceback;
	ViterbiPathLogProb = max;
	 //RECONSTRUCT THE PATH	 					
	bestPathM[totalNumber+1] = endstate;
	i=totalNumber+1;
	bestPathM[i-1] = tracebackM[bestPathM[i]][i];
	for (i=totalNumber; i>=1; i--) 	
		{
		bestPathM[i-1] = tracebackM[bestPathM[i]][i];   
		}
	}
else
	{
	ViterbiPathLogProb = -5000000000.;  //  fictitious, for vit learning protocol
	cerr <<  "\nWarning: Sequence Nb  " << seqNb  << "  has Probability 0"  << endl;
	i=totalNumber+1; tracebackM[s][i] = startstate;
	for (i=totalNumber; i>=1; i--) 	{bestPathM[i-1] = startstate;  }
	}
}
// ---------------------------------------------------------------------------


// ---------------------------------------------------------------------------
//	 Forward-backward algorithm
// --------------------------------------------------------- ------------------
void ModelFB::FBalgorithm (int seqNb, int totalNumber, const int Seq[]) 
{
int	st, pos, s, i, n;
long double probSeq,nb, factor, Rfactor, Cfactor, f, corr; 
bool FBdebugging;
long double intermediate[200]; 
FILE *cFpWrite=NULL;
int expcorr=kExpCorr;
int scalePeriod = kScalePeriod;

if (verbose) cerr << "FB computation STARTS   nbStates = " <<nbStates<<" seqLen="<<totalNumber<< endl;

FBdebugging = false;writeControls = false;
factor = exp(expcorr);  Rfactor = 1 / factor ;  
FBscaleFactor=factor;
for (i=0; i<=totalNumber+1; i++)		 {FBFactor[i] = 0;}

if (writeControls || FBdebugging) 
	{
	if  ( (cFpWrite = fopen("FBalgo.check","w" ))  == NULL )
		{ cerr << "\n could not open the file FBalgo.check for writing \n";}
	else  fprintf(cFpWrite,"\n *********************** \n");
	}

if (writeControls) fprintf(cFpWrite,"\n\n\n  FB computation  nbStates = %d  \n\n\n", nbStates );
for (st=0; st<nbStates; st++) scForw[st][0] = 0.0;
scForw[startstate][0] = 1.0;

if (writeControls)
	{
	for (st=0; st<nbStates; st++) fprintf(cFpWrite,"\n scForw[st %d][pos 0] = %6.15f; ", st, (double)scForw[st][0] );
	for (st=0; st<nbStates; st++) printf("\n scForw[st %d][pos 0] = %6.15f; ", st, (double)scForw[st][0] );
	}

for (pos=1; pos<=totalNumber; pos++)		
	{
	for (st=0; st<nbStates; st++)	
		{
		scForw[st][pos] = 0.0;
		for (n=BackConnFirst[st]; n<=BackConnLast[st]; n++)
			{
			s = BackConnList[n];
			scForw[st][pos] += (scForw[s][pos-1] * scTrProb[s][st]);
			}
		scForw[st][pos] = scForw[st][pos] * scEmProb[st][Seq[pos]]; 
		if (FBdebugging)
			{
			fprintf(cFpWrite, "\n scForw[st %d][pos %d ]   = %6.15f;  ",st,pos,(double) scForw[st][pos] );
		}	}

	if (writeControls)  fprintf(  cFpWrite, "\n scForw[0][pos] = %f  at pos %d  \n" , (double) scForw[0][pos] , pos );
	if (( (pos % scalePeriod)  == 0) && (pos < totalNumber) && (pos > 1)) 	
		{
		nb = scForw[0][pos];  
		for (st=1; st<nbStates; st++)	{ if (scForw[st][pos] > nb)  nb = scForw[st][pos];}
		if ( log(nb) > expcorr )
			{
			for (st=0; st<nbStates; st++)	scForw[st][pos] = scForw[st][pos] * Rfactor;
			FBFactor[pos] = 1;	
			if (writeControls)  fprintf(  cFpWrite, "\n scFBFactor at pos %d  \n", pos );
			}
		else
			{
			if ( log(nb) < -0.5*expcorr )
				{
				for (st=0; st<nbStates; st++)	scForw[st][pos] = scForw[st][pos] * factor;
				FBFactor[pos] = -1;	
				if (writeControls)  fprintf(  cFpWrite, "\n scFBFactor at pos %d  \n", pos );
			}	}
	}	}	

for (st=0; st<nbStates; st++)  scForw[st][totalNumber+1] = 0.0;
for (n=BackConnFirst[endstate]; n <= BackConnLast[endstate] ; n++)
		{
		s = BackConnList[n];
		scForw[endstate][totalNumber+1] += (scForw[s][totalNumber] * scTrProb[s][endstate] ) ;
		}
if (FBdebugging)
	{
	for (st=0; st<nbStates; st++)	
		{
	fprintf(cFpWrite, "\n  scForw[st %d][totalNumber+1 = %d]   = %6.15f;  ", st, totalNumber+1,  (double) scForw[st][totalNumber+1] );
		}
	fprintf(  cFpWrite, "\n\n");
	}

					// BACKWARD
for (s=0; s<nbStates; s++)	 
	{
	scBackw[s][totalNumber+1] = 0.0 ;
	}
scBackw[endstate][totalNumber+1] = 1.0;
if (FBdebugging)
	{
	for (st=0; st<nbStates; st++)	
		{
	fprintf(  cFpWrite, "\n  scBackw[st %d][totalNumber+1 = %d]   = %6.15f;  ", st, totalNumber+1,  (double) scBackw[st][totalNumber+1] );
		}
	}

for (s=0; s<nbStates; s++) 
	{
	scBackw[s][totalNumber] = scTrProb[s][endstate]; 
	intermediate[s] = scBackw[s][totalNumber] *scEmProb[s][Seq[totalNumber]];
	}
if (FBdebugging)
	{
	for (st=0; st<nbStates; st++)	
		{
	fprintf(  cFpWrite, "\n  scBackw[st %d][%d ]   = %6.15f;  ", st,totalNumber , (double)  scBackw[st][totalNumber] );
	}	}

for (pos=totalNumber-1; pos>=1; pos--)		
	{
	if ((FBFactor[pos+1] == 1)	|| (FBFactor[pos+1] == -1))
		{
		if (FBFactor[pos+1] == -1)	f = factor;
		else  f = Rfactor;
		for (st=0; st<nbStates; st++)	
			{
			scBackw[st][pos] = 0.0 ;
			for (n=ForwConnFirst[st]; n <= ForwConnLast[st] ; n++)
				{
				s = ForwConnList[n];
				scBackw[st][pos] += scTrProb[st][s] * intermediate[s] * f ;
		}	}	}
	else	
		{
		for (st=0; st<nbStates; st++)	
			{
			scBackw[st][pos] = 0.0 ;
			for (n=ForwConnFirst[st]; n <= ForwConnLast[st] ; n++)
				{
				s = ForwConnList[n];
				scBackw[st][pos] += scTrProb[st][s] * intermediate[s];
				}
			if (FBdebugging)  printf("\n  scBackw[st %d][%d ]   = %6.15f;  ", st,pos , (double) scBackw[st][pos] );
		}	}	
	for (st=0; st<nbStates; st++)   intermediate[st] = scBackw[st][pos] * scEmProb[st][Seq[pos]];
	}			    

for (s=0; s<nbStates; s++)	scBackw[s][0] = 0.0; 
for (s=0; s<nbStates; s++)	scBackw[startstate][0] += scTrProb[startstate][s] * intermediate[s];

if (writeControls)
      {
	for (s=0; s<nbStates; s++)	 fprintf(cFpWrite, "\n scBackw[s][0] = %6.15f;", (double) scBackw[s][0]);
		
	fprintf(cFpWrite, "\n  scBackw[startstate][0]  = %6.15f;  ;" ,   (double) scBackw[startstate][0]) ;
	fprintf(cFpWrite, "\n  scForw[endstate][totalNumber+1=%d] = %6.15f;  . \n\n\n", totalNumber+1,  (double) scForw[endstate][totalNumber+1] ); 
	}
	
nb = (scBackw[startstate][0] - scForw[endstate][totalNumber+1]) / scBackw[startstate][0];
if (nb < 0.0)  nb = - nb; 
if (nb > kEpsilon)  
	{
	cerr<< "\n seqNb=" << seqNb << "Failed ForwBackw-PrecisionCheck";  	
	cerr<< "\n scBackw[startstate][0] = " << scBackw[startstate][0];  	
	cerr<< "\n scForw[endstate][totalNumber+1] = " << scForw[endstate][totalNumber+1];
	}  	
else
	{
	if (verbose)
		{
		 cerr<< "seqNb=" << seqNb << " Passed ForwBackw-PrecisionCheck at diff <= " << kEpsilon << endl;  	
		cerr<< "\n scBackw[startstate][0] = " << scBackw[startstate][0] << endl;  	
		cerr<< "\n scForw[endstate][totalNumber+1] = " << scForw[endstate][totalNumber+1] << endl;
		}
	}

s=0;
for (pos=0; pos<=totalNumber+1; pos++)
			{
			s += FBFactor[pos];
			if (writeControls) 
				fprintf(cFpWrite, "\n  pos=%d  FBFactor=%d  s=%d", pos, FBFactor[pos], s);
			}
corr = s*log(factor) - totalNumber*log(emScale);

bFBseqProbPositive = true;
if (scBackw[startstate][0] == 0)
	{
	cerr << "\nWarning: Sample with computed FB probability equal to zero" << endl;
	bFBseqProbPositive = false;
	}
else
	{if (verbose) {cerr << "\n Sample with computed FB probability above zero" << endl;}}


logFBseqProb = log(scBackw[startstate][0]) + corr;  
if (verbose) cerr << "log probability  " << logFBseqProb << endl;
if (writeControls)
      {
	fprintf(cFpWrite, "\n\n renormalised prob-score from Backw =  %6.15f;  ;", (double) scBackw[startstate][0]) ;
	fprintf(cFpWrite, "\n renormalised prob-score from Forw =  %6.15f;  ;", (double) scForw[endstate][totalNumber+1]) ;
	fprintf(cFpWrite, "\n corrected prob from Backw =  %6.15f;  ;", (double)  scBackw[startstate][0] * exp(corr));
	fprintf(cFpWrite, "\n corrected log prob from Backw =  %6.15Lf;  ;" ,(double) log(scBackw[startstate][0]) + corr);
	fprintf(cFpWrite, "\n corrected prob from Forw =  %6.15f;  ;", (double) scForw[endstate][totalNumber+1] * exp(corr));
	fprintf(cFpWrite, "\n corrected log prob from Forw =  %6.15Lf; ", (double) log(scForw[endstate][totalNumber+1]) + corr);
	}

// POSTERIOR DECODING , CHECKING	
 Cfactor = 1 / scBackw[startstate][0];
for (pos=0; pos<=totalNumber+1; pos++)
     {
     probSeq = 0.0 ;
	if (FBdebugging)   cout << "\n pos=" << pos<<endl;
     for (st=0; st<nbStates; st++)
	     {
	       PosStateProb[pos][st] = (double) (scForw[st][pos] * scBackw[st][pos] * Cfactor);
		   //used in Baum-Welch , do not remove
 	       probSeq += PosStateProb[pos][st] ;
		  if (FBdebugging)  fprintf (cFpWrite, "\n  PosStateProb[pos][st]  st = %d pos %d = %6.15f;  ",st,  pos,  (double) PosStateProb[pos][st]);
		  }
     if (FBdebugging)   fprintf(cFpWrite, "\n  probSeq at pos %d = %6.15f;  ", pos,  (double) probSeq);
     nb = probSeq - 1.0;
     if (nb < 0.0)  nb = - nb; 
     if (nb > kEpsilon)   cerr  << "\n Failed the probSeq=1 PrecisionCheck in the method ModelFB::FBalgorithm " << endl;  	
     }
if (writeControls || FBdebugging)     fclose( cFpWrite);
if (verbose)  cerr << "FB computation concluded"  << endl;
}
/* FBalgorithm
************************************/

// ---------------------------------------------------------------------------
//	 Forward-backward algorithm  FBalgorithmExtP
// --------------------------------------------------------- ------------------
void ModelFB::FBalgorithmExtP (int seqNb, int totalNumber, const int Seq[]) 
{
int	st, pos, s, i, n;
long double probSeq,nb, factor, Rfactor, Cfactor, f, corr; 
bool FBdebugging;
long double intermediate[200]; 
FILE *cFpWrite=NULL;
int expcorr=kExpCorr;
int scalePeriod = kScalePeriod;

if (verbose) cerr << "FB computation FBalgorithmExtP STARTS nbStates = " <<nbStates<<" seqLen="<<totalNumber<< endl;
FBdebugging = false;writeControls = false;
factor = exp(expcorr);  Rfactor = 1 / factor ;  
FBscaleFactor=factor;
for (i=0; i<=totalNumber+1; i++)		 {FBFactor[i] = 0;}

if (writeControls || FBdebugging) 
	{
	if  ( (cFpWrite = fopen("FBalgo.check","w" ))  == NULL )
		{ cerr << "\n could not open the file FBalgo.check for writing \n";}
	else  fprintf(cFpWrite,"\n *********************** \n");
	}

if (writeControls) fprintf(cFpWrite,"\n\n\n  FB computation  nbStates = %d  \n\n\n", nbStates );

for (st=0; st<nbStates; st++) scForw[st][0] = 0.0;
scForw[startstate][0] = 1.0;

if (writeControls)
	{
	for (st=0; st<nbStates; st++) fprintf(cFpWrite,"\n scForw[st %d][pos 0] = %6.15f; ", st, (double)scForw[st][0] );
	for (st=0; st<nbStates; st++) printf("\n scForw[st %d][pos 0] = %6.15f; ", st, (double)scForw[st][0] );
	}

for (pos=1; pos<=totalNumber; pos++)		
	{
	for (st=0; st<nbStates; st++)	
		{
		scForw[st][pos] = 0.0;
		for (n=BackConnFirst[st]; n<=BackConnLast[st]; n++)
			{
			s = BackConnList[n];
			scForw[st][pos] += (scForw[s][pos-1] * scTrProb[s][st]);
			}
		scForw[st][pos] = scForw[st][pos] * scEmProb[st][Seq[pos]] * extP[st][pos]; 
		if (FBdebugging)
			{
			fprintf(cFpWrite, "\n scForw[st %d][pos %d ]   = %6.15f;  ",st,pos,(double) scForw[st][pos] );
		}	}

		// modified Rabiner-Type overflow-protection / SCALING
	if (writeControls)  fprintf(  cFpWrite, "\n scForw[0][pos] = %f  at pos %d  \n" , (double) scForw[0][pos] , pos );
	if (( (pos % scalePeriod)  == 0) && (pos < totalNumber) && (pos > 1)) 	//only once in scalePeriod  
		{
		nb = scForw[0][pos];  
		for (st=1; st<nbStates; st++)	{ if (scForw[st][pos] > nb)  nb = scForw[st][pos];}
		if ( log(nb) > expcorr )
			{
			for (st=0; st<nbStates; st++)	scForw[st][pos] = scForw[st][pos] * Rfactor;
			FBFactor[pos] = 1;	
			if (writeControls)  fprintf(  cFpWrite, "\n scFBFactor at pos %d  \n", pos );
			}
		else
			{
			if ( log(nb) < -0.5*expcorr )
				{
				for (st=0; st<nbStates; st++)	scForw[st][pos] = scForw[st][pos] * factor;
				FBFactor[pos] = -1;	
				if (writeControls)  fprintf(  cFpWrite, "\n scFBFactor at pos %d  \n", pos );
			}	}
	}	}	

for (st=0; st<nbStates; st++)  scForw[st][totalNumber+1] = 0.0;
for (n=BackConnFirst[endstate]; n <= BackConnLast[endstate] ; n++)
		{
		s = BackConnList[n];
		scForw[endstate][totalNumber+1] += (scForw[s][totalNumber] * scTrProb[s][endstate] ) ;
		}
if (FBdebugging)
	{
	for (st=0; st<nbStates; st++)	
		{
	fprintf(cFpWrite, "\n  scForw[st %d][totalNumber+1 = %d]   = %6.15f;  ", st, totalNumber+1,  (double) scForw[st][totalNumber+1] );
		}
	fprintf(  cFpWrite, "\n\n");
	}

					// BACKWARD
for (s=0; s<nbStates; s++)	 
	{
	scBackw[s][totalNumber+1] = 0.0 ;
	}
scBackw[endstate][totalNumber+1] = 1.0;
if (FBdebugging)
	{
	for (st=0; st<nbStates; st++)	
		{
	fprintf(  cFpWrite, "\n  scBackw[st %d][totalNumber+1 = %d]   = %6.15f;  ", st, totalNumber+1,  (double) scBackw[st][totalNumber+1] );
		}
	}

for (s=0; s<nbStates; s++) 
	{
	scBackw[s][totalNumber] = scTrProb[s][endstate]; 
	intermediate[s] = scBackw[s][totalNumber] *scEmProb[s][Seq[totalNumber]]* extP[s][totalNumber];
	}

if (FBdebugging)
	{
	for (st=0; st<nbStates; st++)	
		{
	fprintf(  cFpWrite, "\n  scBackw[st %d][%d ]   = %6.15f;  ", st,totalNumber , (double)  scBackw[st][totalNumber] );
	}	}

for (pos=totalNumber-1; pos>=1; pos--)		
	{
	if ((FBFactor[pos+1] == 1)	|| (FBFactor[pos+1] == -1))
		{
		if (FBFactor[pos+1] == -1)	f = factor;
		else  f = Rfactor;
		for (st=0; st<nbStates; st++)	
			{
			scBackw[st][pos] = 0.0 ;
			for (n=ForwConnFirst[st]; n <= ForwConnLast[st] ; n++)
				{
				s = ForwConnList[n];
				scBackw[st][pos] += scTrProb[st][s] * intermediate[s] * f ;
		}	}	}
	else	
		{
		for (st=0; st<nbStates; st++)	
			{
			scBackw[st][pos] = 0.0 ;
			for (n=ForwConnFirst[st]; n <= ForwConnLast[st] ; n++)
				{
				s = ForwConnList[n];
				scBackw[st][pos] += scTrProb[st][s] * intermediate[s];
				}
			if (FBdebugging)  printf("\n  scBackw[st %d][%d ]   = %6.15f;  ", st,pos , (double) scBackw[st][pos] );
		}	}	
	for (st=0; st<nbStates; st++)   intermediate[st] = scBackw[st][pos] * scEmProb[st][Seq[pos]]* extP[st][pos];
	}			    

for (s=0; s<nbStates; s++)	scBackw[s][0] = 0.0; 
for (s=0; s<nbStates; s++)	scBackw[startstate][0] += scTrProb[startstate][s] * intermediate[s];

if (writeControls)
      {
	for (s=0; s<nbStates; s++)	 fprintf(cFpWrite, "\n scBackw[s][0] = %6.15f;", (double) scBackw[s][0]);
		
	fprintf(cFpWrite, "\n  scBackw[startstate][0]  = %6.15f;  ;" ,   (double) scBackw[startstate][0]) ;
	fprintf(cFpWrite, "\n  scForw[endstate][totalNumber+1=%d] = %6.15f;  . \n\n\n", totalNumber+1,  (double) scForw[endstate][totalNumber+1] ); 
	}
	
nb = (scBackw[startstate][0] - scForw[endstate][totalNumber+1]) / scBackw[startstate][0];
if (nb < 0.0)  nb = - nb; 
if (nb > kEpsilon)  
	{
	cerr<< "\n seqNb=" << seqNb << "Failed ForwBackw-PrecisionCheck";  	
	cerr<< "\n scBackw[startstate][0] = " << scBackw[startstate][0];  	
	cerr<< "\n scForw[endstate][totalNumber+1] = " << scForw[endstate][totalNumber+1];
	}  	
else
	{
	if (verbose)
		{
		 cerr<< "seqNb=" << seqNb << " Passed ForwBackw-PrecisionCheck at diff <= " << kEpsilon << endl;  	
		cerr<< "\n scBackw[startstate][0] = " << scBackw[startstate][0] << endl;  	
		cerr<< "\n scForw[endstate][totalNumber+1] = " << scForw[endstate][totalNumber+1] << endl;
		}
	}

s=0;
for (pos=0; pos<=totalNumber+1; pos++)
			{
			s += FBFactor[pos];
			if (writeControls) 
				fprintf(cFpWrite, "\n  pos=%d  FBFactor=%d  s=%d", pos, FBFactor[pos], s);
			}
corr = s*log(factor) - totalNumber*log(emScale);

bFBseqProbPositive = true;
if (scBackw[startstate][0] == 0)
	{
	cerr << "\nWarning: Sample with computed FB probability equal to zero" << endl;
	bFBseqProbPositive = false;
	}
else
	{if (verbose) {cerr << "\n Sample with computed FB probability above zero" << endl;}}

logFBseqProb = log(scBackw[startstate][0]) + corr;  
if (verbose) cerr << "log probability  " << logFBseqProb << endl;
if (writeControls)
      {
	fprintf(cFpWrite, "\n\n renormalised prob-score from Backw =  %6.15f;  ;", (double) scBackw[startstate][0]) ;
	fprintf(cFpWrite, "\n renormalised prob-score from Forw =  %6.15f;  ;", (double) scForw[endstate][totalNumber+1]) ;
	fprintf(cFpWrite, "\n corrected prob from Backw =  %6.15f;  ;", (double)  scBackw[startstate][0] * exp(corr));
	fprintf(cFpWrite, "\n corrected log prob from Backw =  %6.15Lf;  ;" ,(double) log(scBackw[startstate][0]) + corr);
	fprintf(cFpWrite, "\n corrected prob from Forw =  %6.15f;  ;", (double) scForw[endstate][totalNumber+1] * exp(corr));
	fprintf(cFpWrite, "\n corrected log prob from Forw =  %6.15Lf; ", (double) log(scForw[endstate][totalNumber+1]) + corr);
	}

// POSTERIOR DECODING , CHECKING	
 Cfactor = 1 / scBackw[startstate][0];
for (pos=0; pos<=totalNumber+1; pos++)
     {
     probSeq = 0.0 ;
	if (FBdebugging)   cout << "\n pos=" << pos<<endl;
     for (st=0; st<nbStates; st++)
	     {
	       PosStateProb[pos][st] = (double) (scForw[st][pos] * scBackw[st][pos] * Cfactor);
		   //used in Baum-Welch , do not remove
 	       probSeq += PosStateProb[pos][st] ;
		  if (FBdebugging)  fprintf (cFpWrite, "\n  PosStateProb[pos][st]  st = %d pos %d = %6.15f;  ",st,  pos,  (double) PosStateProb[pos][st]);
		  }
     if (FBdebugging)   fprintf(cFpWrite, "\n  probSeq at pos %d = %6.15f;  ", pos,  (double) probSeq);
     nb = probSeq - 1.0;
     if (nb < 0.0)  nb = - nb; 
     if (nb > kEpsilon)   cerr  << "\n Failed the probSeq=1 PrecisionCheck in the method ModelFB::FBalgorithm " << endl;  	
     }
if (writeControls || FBdebugging)     fclose( cFpWrite);
if (verbose)  cerr << "FB computation FBalgorithmExtP concluded"  << endl;
}
/* FBalgorithmExtP
************************************/


// ---------------------------------------------------------------------------
//		 InitCodes 					
// ---------------------------------------------------------------------------
void	ModelFB::InitCodes(void)
{
int i, t;

if (verbose)   cerr << "\nstart in initCodes in class ModelFB";
for (t=1; t<=nbLetters;  t++)  Code[(int)(letter[t])] = t;

//for historical reason keeping Decode
for (i=1; i<=nbLetters; i++)	 Decode[i] = (int) letter[i];
Decode[0] = 24 + kAtoOne;

if (verbose)
	{
	for (t=1; t<=nbLetters;  t++)  
		printf("\n t=%d  letter=%c  code=%d", t , letter[t] , Code[(int)(letter[t])] );

	for (t=1; t<=nbLetters;  t++)  
		printf("\n t=%d  Decode=%c ", t , Decode[t]);
	printf("\n ");
	}

if (verbose) cerr << "finished in initCodes"<< endl;
}
/* init.codes
************************************/

// ---------------------------------------------------------------------------
//	ComputeSeqBW, after FB
// ---------------------------------------------------------------------------
void	ModelFB::ComputeSeqBW(int seqNb, int totalNumber, const int Seq[])
{
int i,k, t,j,st, pos;
float weight;
long double factor, nb;
if (verbose) cerr << "ModelFB::ComputeSeqBW" << endl;

for (i=0; i<nbStates; i++)	
	{
	for (k=1; k<=nbLetters ; k++)  Cont[i][k] = 0; 
	for (k=0; k<nbStates; k++)     tCont[i][k] = 0; 
	}

// compute and sum the BAUM-WELCH contributions from this sequence
pos=0;
factor = 1;
if (FBFactor[pos+1] == -1)	factor = FBscaleFactor;
if (FBFactor[pos+1] == 1)	factor = 1 / FBscaleFactor;

if (scBackw[startstate][0] > 0)
	{
	for (st=0; st<nbStates; st++)	 
		{
		for (i=0; i<nbStates; i++) 
			{
			nb=scForw[st][0]*scBackw[i][1]*factor*scTrProb[st][i]*scEmProb[i][Seq[1]]/scBackw[startstate][0];
			tCont[st][i]+=nb;
			}
		}

	for (pos=1; pos<=totalNumber; pos++)		
		{
		factor = 1;
		if (FBFactor[pos+1] == -1)	factor = FBscaleFactor;
		if (FBFactor[pos+1] == 1)	factor = 1 / FBscaleFactor;
		for (st=0; st<nbStates; st++)	 
			{
			nb = PosStateProb[pos][st];
			Cont[st][Seq[pos]] +=  PosStateProb[pos][st]; 
			for (i=0; i<nbStates; i++) 
				{
				if (pos==totalNumber)  
				nb=scForw[st][pos]*scBackw[i][pos+1]*factor*scTrProb[st][i]/scBackw[startstate][0];
				else
				nb=scForw[st][pos]*scBackw[i][pos+1]*factor*scTrProb[st][i]*scEmProb[i][Seq[pos+1]]/scBackw[startstate][0];
				tCont[st][i]+=nb;
				}
			}
		}
	}
else		
	{;} 

if (busecties)
	{
	for (t=1; t<=nbcEMties; t++) 
		{ 
		for (k=1; k<=4 ; k++) // A C G T = 1 2 3 4 
			{
			Cont[cties[t][0]][k] += Cont[cties[t][1]][5-k];
			}
		for (k=1; k<=4 ; k++)  
			{
			Cont[cties[t][1]][k] = Cont[cties[t][0]][5-k];
			}
		}	
	}
			
if (buseties)
	{
	for (t=1; t<=nbEMties; t++) // for all groups 
		{ 
		for (j=2; j<=ties[t][0]; j++) // donor state index
			{
			for (k=1; k<=nbLetters ; k++)  
				{
				Cont[ties[t][1]][k] += Cont[ties[t][j]][k];
				// every letter of every state in the groups receives also
				//the contributions from every other state in the group
				}
			}
		for (j=2; j<=ties[t][0]; j++) 
			{
			for (k=1; k<=nbLetters ; k++)  
				{
				Cont[ties[t][j]][k] = Cont[ties[t][1]][k];
				}
			}
		}	
	}

weight=1;
if (bupdatetrans)
	{
	for (st=0; st<nbStates; st++)
		{ 
		for (i=0; i<nbStates; i++) 	
			{
			scSumTrTot[st][i] += weight * tCont[st][i] ; 
			}
		}
	}
	
if (bupdateemiss)
	{
	for (st=0; st<nbStates; st++)
		{ 
		for (k=1; k<=nbLetters; k++)  
			{
			scSumEmTot[st][k] += weight * Cont[st][k]; 
			}
		}
	}
}
/* ComputeSeqBW
************************************/

// ---------------------------------------------------------------------------
//	ComputeSeqBWdsDNA1, after FB ds DNA models 
// ---------------------------------------------------------------------------
void	ModelFB::ComputeSeqBWdsDNA1(int seqNb, int totalNumber, const int Seq[])
{
int i,k,j,t, st, pos;
long double factor, nb;
if (verbose) cerr << "start ModelFB::ComputeSeqBWdsDNA1" << endl;

for (i=0; i<nbStates; i++)	
	{
	for (k=1; k<=nbLetters ; k++)  Cont[i][k] = 0; 
	for (k=0; k<nbStates; k++)     tCont[i][k] = 0; 
	}

pos=0;
factor = 1;
if (FBFactor[pos+1] == -1)	factor = FBscaleFactor;
if (FBFactor[pos+1] == 1)	factor = 1 / FBscaleFactor;

if (scBackw[startstate][0] > 0)
	{
	for (st=0; st<nbStates; st++)	 
		{
		for (i=0; i<nbStates; i++) 
			{
			nb=scForw[st][0]*scBackw[i][1]*factor*scTrProb[st][i]*scEmProb[i][Seq[1]]/scBackw[startstate][0];
			tCont[st][i]+=nb;
			}
		}

	for (pos=1; pos<=totalNumber; pos++)		
		{
		factor = 1;
		if (FBFactor[pos+1] == -1)	factor = FBscaleFactor;
		if (FBFactor[pos+1] == 1)	factor = 1 / FBscaleFactor;
		for (st=0; st<nbStates; st++)	 
			{
			nb = PosStateProb[pos][st];
			Cont[st][Seq[pos]] +=  PosStateProb[pos][st]; 
			for (i=0; i<nbStates; i++) 
				{
				if (pos==totalNumber)  
				nb=scForw[st][pos]*scBackw[i][pos+1]*factor*scTrProb[st][i]/scBackw[startstate][0];
				else
				nb=scForw[st][pos]*scBackw[i][pos+1]*factor*scTrProb[st][i]*scEmProb[i][Seq[pos+1]]/scBackw[startstate][0];
				tCont[st][i]+=nb;
				}
			}
		}

	if (busecties)
		{
		for (t=1; t<=nbcEMties; t++) // for all pairs 
			{ 
			cerr << "t  = " << t << endl;
			cerr << "first state in pair  = " << cties[t][0] << endl;
			cerr << "second state in pair  = " << cties[t][1] << endl;
			for (k=1; k<=4 ; k++) // A C G T = 1 2 3 4 
				{
				Cont[cties[t][0]][k] += Cont[cties[t][1]][5-k];
			cerr << "Cont.A 100k+cties[t][0] = " <<   100*k+cties[t][0] << "  = " << Cont[cties[t][0]][k]  << endl;
			cerr << "Cont.A 1 = " <<   Cont[cties[t][0]][k]  << endl;
			cerr << "Cont.A 2 = " <<   Cont[cties[t][1]][5-k]  << endl;
				}
			for (k=1; k<=4 ; k++)  
				{
				Cont[cties[t][1]][k] = Cont[cties[t][0]][5-k];
			cerr << "Cont.B 100k+cties[t][1] = " <<   100*k+cties[t][1] << "  = " << Cont[cties[t][1]][k]  << endl;
				}
			}	
		}


	if (buseties)
		{
		for (t=1; t<=nbEMties; t++) 
			{ 
			for (j=2; j<=ties[t][0]; j++) // donor state index 
				{
				for (k=1; k<=nbLetters ; k++)  
					{
					Cont[ties[t][1]][k] += Cont[ties[t][j]][k];
					// every letter of every state in the groups receives also
					//the contributions from every other state in the group
					}
				}
			for (j=2; j<=ties[t][0]; j++) 
				{
				for (k=1; k<=nbLetters ; k++)  
					{
					Cont[ties[t][j]][k] = Cont[ties[t][1]][k];
					}
				}
			}	
		}
	}
else	{;} 

if (verbose) cerr << "finish ModelFB::ComputeSeqBWdsDNA1" << endl;
}
/* ComputeSeqBW
************************************/

// ---------------------------------------------------------------------------
//	ComputeSeqBWdsDNA2, after FB ds DNA models 
// ---------------------------------------------------------------------------
void	ModelFB::ComputeSeqBWdsDNA2(int seqNb, int totalNumber, const int Seq[], long double secstrandweight )
{
int i,k, st, t, j, pos;
long double factor, nb;
if (verbose) cerr << "start ModelFB::ComputeSeqBWdsDNA2" << endl;

for (i=0; i<nbStates; i++)	
	{
	for (k=1; k<=nbLetters ; k++)  Cont2[i][k] = 0; 
	for (k=0; k<nbStates; k++)     tCont2[i][k] = 0; 
	}

pos=0;
factor = 1;
if (FBFactor[pos+1] == -1)	factor = FBscaleFactor;
if (FBFactor[pos+1] == 1)	factor = 1 / FBscaleFactor;

if (scBackw[startstate][0] > 0)
	{
	for (st=0; st<nbStates; st++)	 
		{
		for (i=0; i<nbStates; i++) 
			{
			nb=scForw[st][0]*scBackw[i][1]*factor*scTrProb[st][i]*scEmProb[i][Seq[1]]/scBackw[startstate][0];
			tCont2[st][i]+=nb;
			}
		}

	for (pos=1; pos<=totalNumber; pos++)		
		{
		factor = 1;
		if (FBFactor[pos+1] == -1)	factor = FBscaleFactor;
		if (FBFactor[pos+1] == 1)	factor = 1 / FBscaleFactor;
		for (st=0; st<nbStates; st++)	 
			{
			nb = PosStateProb[pos][st];
			Cont2[st][Seq[pos]] +=  PosStateProb[pos][st]; 
			for (i=0; i<nbStates; i++) 
				{
				if (pos==totalNumber)  
				nb=scForw[st][pos]*scBackw[i][pos+1]*factor*scTrProb[st][i]/scBackw[startstate][0];
				else
				nb=scForw[st][pos]*scBackw[i][pos+1]*factor*scTrProb[st][i]*scEmProb[i][Seq[pos+1]]/scBackw[startstate][0];
				tCont2[st][i]+=nb;
				}
			}
		}
	}
else		
	{;} 

if (busecties)
	{
	for (t=1; t<=nbcEMties; t++) 
		{ 
			cerr << "t  = " << t << endl;
			cerr << "first state in pair  = " << cties[t][0] << endl;
			cerr << "second state in pair  = " << cties[t][1] << endl;
		for (k=1; k<=4 ; k++) // A C G T = 1 2 3 4 
			{
			Cont2[cties[t][0]][k] += Cont2[cties[t][1]][5-k];
			cerr << "Cont2.A 100k+cties[t][0] = " <<   100*k+cties[t][0] << "  = " << Cont2[cties[t][0]][k]  << endl;
			}
		for (k=1; k<=4 ; k++)  
			{
			Cont2[cties[t][1]][k] = Cont2[cties[t][0]][5-k];
			cerr << "Cont2.B 100k+cties[t][1] = " <<   100*k+cties[t][1] << "  = " << Cont2[cties[t][1]][k]  << endl;
			}
		}	
	}
			
if (buseties)
	{
	for (t=1; t<=nbEMties; t++) 
		{ 
		for (j=2; j<=ties[t][0]; j++) 
			{
			for (k=1; k<=nbLetters ; k++)  
				{
				Cont2[ties[t][1]][k] += Cont2[ties[t][j]][k];
				}
			}
		for (j=2; j<=ties[t][0]; j++) 
			{
			for (k=1; k<=nbLetters ; k++)  
				{
				Cont2[ties[t][j]][k] = Cont2[ties[t][1]][k];
				}
			}
		}	
	}

if (bupdatetrans)
	{
	for (st=0; st<nbStates; st++)
		{ 
		for (i=0; i<nbStates; i++) 	
			{
			scSumTrTot[st][i] += (1-secstrandweight) * tCont[st][i] + secstrandweight * tCont2[st][i];
			}
		}
	}
	
if (bupdateemiss)
	{
	for (st=0; st<nbStates; st++)
		{ 
		for (k=1; k<=nbLetters; k++)  
			{
			scSumEmTot[st][k] += (1-secstrandweight) * Cont[st][k] + secstrandweight * Cont2[st][k];
			 // with strand specific weight factor 
			}
		}
	}

}
/* ComputeSeqBW
************************************/

// ---------------------------------------------------------------------------
//	ComputeSeqVL, after VITERBI	standard 	
// ---------------------------------------------------------------------------
void	ModelFB::ComputeSeqVL(int seqNb, int totalNumber, const int Seq[])
{
int i,k, st, pos, t, j;
float weight;

if (verbose) cerr << "start ModelFB::ComputeSeqVL" << endl;

for (i=0; i<nbStates; i++)	
	{
	for (k=0; k<=nbLetters ; k++)  Cont[i][k] = 0; 
	for (k=0; k<nbStates; k++)     tCont[i][k] = 0; 
	}

pos=0;

if (possible)
	{
	tCont[bestPathM[0]][bestPathM[1]]++;

	for (pos=1; pos<=totalNumber; pos++)		
		{
		tCont[bestPathM[pos]][bestPathM[pos+1]]++;
		Cont[bestPathM[pos]][Seq[pos]]++;
		}
	}
else		
	{;} 

if (busecties)
	{
	for (t=1; t<=nbcEMties; t++) 
		{ 
		for (k=1; k<=4 ; k++) // A C G T = 1 2 3 4 
			{
			Cont[cties[t][0]][k] += Cont[cties[t][1]][5-k];
			}
		for (k=1; k<=4 ; k++)  
			{
			Cont[cties[t][1]][k] = Cont[cties[t][0]][5-k];
			}
		}	
	}
			
if (buseties)
	{
	for (t=1; t<=nbEMties; t++) // for all groups 
		{ 
		for (j=2; j<=ties[t][0]; j++) // donor state index
			{
			for (k=1; k<=nbLetters ; k++)  
				{
				Cont[ties[t][1]][k] += Cont[ties[t][j]][k];
				// every letter of every state in the groups receives also
				//the contributions from every other state in the group
				}
			}
		for (j=2; j<=ties[t][0]; j++) 
			{
			for (k=1; k<=nbLetters ; k++)  
				{
				Cont[ties[t][j]][k] = Cont[ties[t][1]][k];
				}
			}
		}	
	}

weight=1;
if (bupdatetrans)
	{
	for (st=0; st<nbStates; st++)
		{ 
		for (i=0; i<nbStates; i++) 	
			{
			scSumTrTot[st][i] += weight * tCont[st][i] ; 
			}
		}
	}
	
if (bupdateemiss)
	{
	for (st=0; st<nbStates; st++)
		{ 
		for (k=1; k<=nbLetters; k++)  
			{
			scSumEmTot[st][k] += weight * Cont[st][k]; 
			}
		}
	}
if (verbose) cerr << "end ModelFB::ComputeSeqVL" << endl;
}
/* ComputeSeqVL
************************************/

// ---------------------------------------------------------------------------
//	ComputeSeqVL, after VITERBI	for ds DNA models 	after first strand 
// ---------------------------------------------------------------------------
void	ModelFB::ComputeSeqVLdsDNA1(int seqNb, int totalNumber, const int Seq[])
{
int i,k,pos;

if (verbose) cerr << "start ModelFB::ComputeSeqVL" << endl;

for (i=0; i<nbStates; i++)	
	{
	for (k=0; k<=nbLetters ; k++)  Cont[i][k] = 0; 
	for (k=0; k<nbStates; k++)     tCont[i][k] = 0; 
	}

pos=0;

if (possible)
	{
	tCont[bestPathM[0]][bestPathM[1]]++;

	for (pos=1; pos<=totalNumber; pos++)		
		{
		tCont[bestPathM[pos]][bestPathM[pos+1]]++;
		Cont[bestPathM[pos]][Seq[pos]]++;
		}
	}
else		
	{;} 

if (verbose) cerr << "end ModelFB::ComputeSeqVLdsDNA1" << endl;
}
/* ComputeSeqVL
************************************/

// ---------------------------------------------------------------------------
//	ComputeSeqVL, after VITERBI	for ds DNA models 	after second strand 	
// ---------------------------------------------------------------------------
void	ModelFB::ComputeSeqVLdsDNA2(int seqNb, int totalNumber, const int Seq[], long double secstrandweight)
{
int i,k,t,j, st, pos;

if (verbose) cerr << "start ModelFB::ComputeSeqVL" << endl;

for (i=0; i<nbStates; i++)	
	{
	for (k=0; k<=nbLetters ; k++)  Cont2[i][k] = 0; 
	for (k=0; k<nbStates; k++)     tCont2[i][k] = 0; 
	}

pos=0;

if (possible)
	{
	tCont2[bestPathM[0]][bestPathM[1]]++;

	for (pos=1; pos<=totalNumber; pos++)		
		{
		tCont2[bestPathM[pos]][bestPathM[pos+1]]++;
		Cont2[bestPathM[pos]][Seq[pos]]++;
		}
	}
else		
	{;} 

if (busecties)
	{
	for (t=1; t<=nbcEMties; t++) 
		{ 
		for (k=1; k<=4 ; k++) // A C G T = 1 2 3 4 
			{
			Cont[cties[t][0]][k] += Cont[cties[t][1]][5-k];
			}
		for (k=1; k<=4 ; k++)  
			{
			Cont[cties[t][1]][k] = Cont[cties[t][0]][5-k];
			}
		}	
	}
			
if (buseties)
	{
	for (t=1; t<=nbEMties; t++) // for all groups 
		{ 
		for (j=2; j<=ties[t][0]; j++) // donor state index
			{
			for (k=1; k<=nbLetters ; k++)  
				{
				Cont[ties[t][1]][k] += Cont[ties[t][j]][k];
				// every letter of every state in the groups receives also
				//the contributions from every other state in the group
				}
			}
		for (j=2; j<=ties[t][0]; j++) 
			{
			for (k=1; k<=nbLetters ; k++)  
				{
				Cont[ties[t][j]][k] = Cont[ties[t][1]][k];
				}
			}
		}	
	}
////////////////////////////////

if (bupdatetrans)
	{
	for (st=0; st<nbStates; st++)
		{ 
		for (i=0; i<nbStates; i++) 	
			{
			scSumTrTot[st][i] += (1-secstrandweight) * tCont[st][i] + secstrandweight * tCont2[st][i]; 
			}
		}
	}
	
if (bupdateemiss)
	{
	for (st=0; st<nbStates; st++)
		{ 
		for (k=1; k<=nbLetters; k++)  
			{
			scSumEmTot[st][k] += (1-secstrandweight) * Cont[st][k] + secstrandweight * Cont2[st][k];
			}
		}
	}
if (verbose) cerr << "end ModelFB::ComputeSeqVLdsDNA2" << endl;
}
/* ComputeSeqVL
************************************/

// ---------------------------------------------------------------------------
//		InitialiseBWVitIterStep 	USED FOR BW AND VITERBI			
// ---------------------------------------------------------------------------
void	ModelFB::InitialiseBWIterStep( void ) 
{	
 int	  s, st, k;
if (verbose) cerr <<"\nstart ModelFB::InitialiseBWIterStep" << endl;
for (st=0; st<nbStates; st++)		
	{
	for (k=1; k<=nbLetters; k++)  scSumEmTot[st][k] = 0;
	for (s=0; s<nbStates; s++)	scSumTrTot[st][s] = 0;
	 }
if (verbose) cerr <<"end ModelFB::InitialiseBWIterStep" << endl;
}
/* InitialiseBWVitIterStep
************************************/

// ---------------------------------------------------------------------------
//		ParameterIterationBW				
// ---------------------------------------------------------------------------
void	ModelFB::ParameterIterationBW( void )
{
int	s, st, k;
long double sum, f;

if (bupdatetrans)
	{
	for (st=0; st<nbStates; st++)		
		{
		if (updatetr[st]==1)
			{
			// 	(1-kTransRet)  scSumTrTot[0][s] + (kTransRet) *  ...
			if (pseudow[st]==1) // ==0 for startstate and endstate
				{
				sum= 0;
				for (s=0; s<nbStates; s++)	
					{
					if (pseudow[s]==1) 
					{scSumTrTot[st][s]=scSumTrTot[st][s] + vpseudoTrans;}
					else
					{scSumTrTot[st][s]=scSumTrTot[st][s];}
					sum+= scSumTrTot[st][s]; 
					}
				if (sum>0) f=1/sum; else f=0;
				for (s=0; s<nbStates; s++)	 trProb[st][s]=scSumTrTot[st][s] * f;
				}	
			else
				{
				sum= 0;
				for (s=0; s<nbStates; s++)	
					{
					sum+= scSumTrTot[st][s]; 
					}
				if (sum>0) f=1/sum; else f=0;
				for (s=0; s<nbStates; s++)	 trProb[st][s]=scSumTrTot[st][s] * f;
				}	
			}
		}
	}

if (bupdateemiss) 
	{
	for (st=0; st<nbStates; st++)		
		{
		if (updateem[st]==1) 
			{
			sum= 0;
			for (k=1; k<=nbLetters; k++)	
				{
				scSumEmTot[st][k]=scSumEmTot[st][k] + pseudocountsEm[k];
				sum+= scSumEmTot[st][k] ;
				}
			if (sum>0) f=1/sum; else f=0;
			for (k=1; k<=nbLetters; k++)	emProb[st][k]= scSumEmTot[st][k] * f;
			}
		}
	}
}
/* ParameterIterationBWpc
************************************/

// ---------------------------------------------------------------------------
//		Pseudocounts				
// ---------------------------------------------------------------------------
void	ModelFB::Pseudocounts( void )
{
int k;

if (verbose) cerr << "\nstart ModelFB::Pseudocounts" << endl;

	for (k=1; k<=nbLetters; k++)	
		{
		pseudocountsEm[k] = pseudocountweight * nullEmFreq[k] * nbLetters;
		}	
if (verbose) cerr << "finished ModelFB::Pseudocounts" << endl;
}
/* ParameterIterationBW
************************************/

// ---------------------------------------------------------------------------
//		ran2 //long period generator of L'Ecuyer (numerical recipes, chap7)
// ---------------------------------------------------------------------------
float	ModelFB::ran2(long &idum )
//returns a uniform (0,1) number >0 and <1
//call with idum a negative interger to initalize, thereafter 
//RNMX should approximate the largest float <1 in machine precision
{
int	j;
long k;
float temp;
static long iy=0;
static long idum2=123456789;
static long iv[NTAB];

//initialise
if (idum <= 0 || !iy) 
	{ 
	if (-(idum) < 1) idum=1; //prevent idum=0;
	else idum = -(idum);
	for (j=NTAB+7;j>=0;j--)//load a shuffling table after 8 warm-ups
		{
		k=idum/IQ1;
		idum=IA1*(idum-k*IQ1)-IR1*k;
		if (idum<0) idum+=IM1;
		if (j<NTAB) iv[j]=idum;//fill table
		}
	iy=iv[0];
	}

k=idum/IQ1;
idum=IA1*(idum-k*IQ1)-IR1*k;//Schrage's method
if (idum<0) idum+=IM1;
k=idum2/IQ2;
idum2=IA2*(idum2-k*IQ2)-k*IR2;
if (idum2<0) idum2+=IM2;

j=iy/NDIV;//range 0..NTAB-1
iy=iv[j]-idum2;
iv[j]=idum;
if (iy < 1) iy+=IMM1;
temp=AM*iy;
if ((temp=AM*iy)>RNMX) 
	{
	return RNMX;//avoid 0,1
	}
else  
	{
	return temp;
	}
}
/* ran2 
************************************/

// ---------------------------------------------------------------------------
//		rcseq				
// ---------------------------------------------------------------------------
void ModelFB::rcseq(int totalNumber, int Seq[]) 
{
int pos;
for (pos=1; pos<=totalNumber; pos++)	{SecondStrand[pos] = 5 - Seq[totalNumber+1-pos];}	
}
/* rcseq
************************************/

