/***********************************************
************************************************
**                                            **
** mamot                                      **
** a program for HMM modelling                **
** 2008, Mauro C. Delorenzi, Frédéric Schütz  **
**                                            **
************************************************
************************************************

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
int vMAXnbITERATIONS;
int vMINnbITERATIONS;
double pseudocountweight, vpseudoTrans;
int nbseqs, maxlen;
long seedran2=1;  // Seed for the random number generator
float vMINdifftotLogLik, vMINProbPrint;
#include "pgm.h"  

// for use with GNU C and getopt #include <getopt.h>
bool bfileoutput, alloutput, storeseqs;

ModelFB model;
void usage() {
  /* Print help info and exit.  */

  /* Note (from GNU hello): if no variable is to be printed, use
     fputs("...", stdout) rather than printf. According to some
     web pages, it is 40% faster */
  printf("mamot -- Markov Modeling Tool (revision %s).\n", SVN_REV);
  printf("Copyright (C) 2008 Swiss Institute of Bioinformatics\n");
  printf ("\n");
  printf ("Usage: mamot [COMMAND] [OPTIONS]\n");
  printf ("\n");
// Long options: only if GNU getopt is used
//  -h, --help          display this help and exit\n
//  -v, --version       display version information and exit\n");
  printf ("\
   -h                  display this help and exit\n\
   -v                  verbose output\n\
\n\
Commands:\n\
\n\
   -G     Generation of random sequences\n\
\n\
   -B     Baum-Welch (BW, EM) Learning\n\
   -V     Viterbi Learning\n\
\n\
   -P     FB (Forward(-backward)) Probability\n\
   -D     FB forward-backward \"posterior\" Decoding\n\
   -Q     Viterbi probability and decoding\n");

   

  printf ("\n");

  // Or use PACKAGE_BUGREPORT
  printf ("Report bugs to <Frederic.Schutz@isb-sib.ch> or <Mauro.Delorenzi@isb-sib.ch>.\n");

  exit (0);
}

// ---------------------------------------------------------------------------
//		 main 					
// ---------------------------------------------------------------------------
int main(int argc, char *argv[])
{
char  *P;
int   i, n, let, len, Number;
char  *ModelFile, *seqFile;
char  optSeqFile[101], optModelFile[101], extPFileName[101];
char  thisletter;
bool  gotSeqFile;

// For GNU C GetOpt
// static struct option const long_options[] =
//   {
//     //    {"number", no_argument, NULL, 'n'},
//     {"help", no_argument, 0, 'h'},
//     {"version", no_argument, 0, 'v'},
//     {NULL, 0, NULL, 0}
//   };

vMINProbPrint = kMINProbPrint;
vMAXnbITERATIONS= kMAXnbITERATIONS;vMINnbITERATIONS = kMINnbITERATIONS;//defaults
vpseudoTrans=kpseudoTrans;vMINdifftotLogLik= kMINdifftotLogLik;
Number = 1;//default:generate one sequence
bool  gotModelFile;
int modus=0;
bfileoutput=false;alloutput=false;storeseqs=false;bstateweights=false;
verbose=false;writeControls=false;bdsDNA=false;bdsDNAjointly=false;buseties=false;
gotSeqFile=false;gotModelFile=false;busecties=false;
bupdatetrans=true; bupdateemiss=true;bUseExtP=false;
nbseqs=maxlen=0;
 
CutoffForWriting = -1.0;DomainThreshold=-1.0;pseudocountweight=0.0;
FBDdetailsC=false;

for (i = 1; i < argc; i++)   
   {
     char *optionFlags = "BDGPQVabcfhptuvx";// - dgijklmnsw  and   + cdlmdS   used below
     switch  (argv [i][0])
       {
       case  '-' :
	 if (strpbrk ( argv [i] , optionFlags)!= NULL )
	   {
	     len = strlen(argv [i]);
	     for (let = 1; let < len; let++) 
	       { 
		 thisletter = argv[i][let];  
		 switch  (thisletter)
		   {	
		   case  'v' :  verbose = true;      		break;
		   case  'a' :  alloutput = true;          break;
		   case  'b' :  bupdatetrans = false;      break;
		   case  'c' :  bupdateemiss = false;      break;
		   case  'e' :  bdsDNAjointly = true;
		     cerr <<  "set option e: assume input is ds DNA and compute both strands conjointly"<< endl;      break;
		   case  'f' :  bfileoutput = true;        break;
		   case  'h' :  usage(); break;
		   case  'p' :  busecties = true;
		     cerr <<  "set option p: use tying of emission probabilities for pairs of complementary states as specified in the model file (only for Baum Welch)"<< endl;      break;
		   case  'u' :  bdsDNA = true;
		     cerr <<  "set option u: assume input is ds DNA and compute both strands independently from each other"<< endl;      break;
		   case  't' :  buseties = true;
		     cerr <<  "set option t: use tying of emission probabilities as specified in the model file (only for Baum Welch)"<< endl;      break;
		   case  'x' :  bstateweights = true;
		     cerr <<  "set option x: use weights of emission probabilities for the states specified in the model file (only for decoding)"<< endl;      break;
		   case  'G' :  modus=1; break;
		   case  'B' :  modus=2; break;
		   case  'D' :  modus=3; FBDdetailsC=true; break;
		   case  'Q' :  modus=4; break;
		   case  'V' :  modus=5; break;
		   case  'P' :  modus=0; FBDdetailsC=false; break;
		   default : cerr <<  "Unrecognized option:  " <<  argv [i] << endl; break;
		   }
	       }
	   }
	 else {
	   switch  (argv [i][1])//smgdniklw
	     {
	     case  's' : 
	       i++;
	       if(i >= argc)   {cerr << "\nError:missing filename after option -s \n";exit(1);} 
	       else  {n = strlen(argv[i]);
		 if (n > 100) {cerr << "\n Warning: filename too long max = 100 char \n"; exit(1);}
		 strncpy(optSeqFile,argv[i], 100);
		 gotSeqFile = true;
	       }
	       break;

	     case  'z' : 
	       i++;
		   cerr <<  "set option z: use additional external position-specific state probabilities"<< endl;  
	       if(i >= argc)   {cerr << "\nError:missing filename after option -z \n";exit(1);} 
	       else  {n = strlen(argv[i]);
		   if (n > 100) {cerr << "\n Warning: filename too long max = 100 char \n"; exit(1);}
		   strncpy(extPFileName,argv[i], 100);
		   bUseExtP= true;extPFile = extPFileName; 
		    }
		break;

	     case  'm' : 
	       i++;
	       if(i >= argc)   {cerr << "\nError:missing filename after option -m \n";exit(1);} 
	       else  {n = strlen(argv[i]);
		 if (n > 100) {cerr << "\n Warning: filename too long max = 100 char \n"; exit(1);}
		 strncpy(optModelFile,argv[i], 100);
		 gotModelFile = true;
		}
		break;
		  
	      case  'g' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2; }// if no space 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -g \n";exit(1);} 
		  else  P = argv [i];  
		} 
		vMINProbPrint = (float) atof (P);
		break;
		  
	      case  'd' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2; } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -d \n";exit(1);} 
		  else  P = argv [i];  
		} 
		vMINdifftotLogLik = (float) atof (P);
		break;

	      case  'n' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2; } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -n \n";exit(1);} 
		  else  P = argv [i];  
		} 
		Number = atoi (P);
		break;

	      case  'r' :  
		if  (argv [i][2] != '\0'){  P = argv [i] + 2; } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -r \n";exit(1);} 
		  else  P = argv [i];  
		} 
		seedran2 = atoi (P);
		break;

	      case  'i' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2;  } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -i \n";exit(1);} 
		  else  P = argv [i];  
		} 
		vMAXnbITERATIONS = atoi (P);
		break;

	      case  'j' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2;  } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -j \n";exit(1);} 
		  else  P = argv [i];  
		} 
		vMINnbITERATIONS = atoi (P);
		break;

/* 
	      case  'k' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2;  } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -k \n";exit(1);} 
		  else  P = argv [i]; //did not bring improvements, stopped supporting this option   
		} 
		nbseqs = atoi (P);  storeseqs = true;   cerr << "\n set storeseqs \n";
		break;
*/
 
	      case  'l' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2;  } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -l \n";exit(1);} 
		  else  P = argv [i];  
		} 
		maxlen = atoi (P);
		break;
 
	      case  'w' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2;  } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -w \n";exit(1);} 
		  else  P = argv [i];  
		} 
		pseudocountweight = (double) atof (P); 
		cerr <<  "set option w: weight for pseudocounts = "<<  pseudocountweight<< endl; 
		break;

	      case  'y' :
		if  (argv [i][2] != '\0'){  P = argv [i] + 2;  } 
		else	{
		  i++;
		  if(i >= argc)   {cerr << "\nError:missing number after option -y \n";exit(1);} 
		  else  P = argv [i];  
		}
		vpseudoTrans = (double) atof (P); 
		cerr <<  "set option y: transitions-pseudocounts = "<< vpseudoTrans  << endl; 
		break;
 
	      default : cerr <<  "Unrecognized option:  " <<  argv [i]  << endl;break;

	      }	} 
	  break;
       case  '+' :
	 len = strlen(argv [i]);
	 for (let = 1; let < len; let++) 
	   { 
	     thisletter = argv [i][let];

	     switch  (thisletter)
	       {	
//	       case  'm' :  FBDdetailsC = false;  
//		   		if (verbose) {cerr << "\n +m => set old-style list of probabilities" << endl;}
//			 	break;
	       case  'c' : 
		 if  (argv [i][2] != '\0'){  P = argv [i] + 2; 
		 } 
		 else	{
		   i++;
		   if(i >= argc)   {cerr << "\nError:missing number after option +c \n";exit(1);} 
		   else  P = argv [i];  
		 } 
		 CutoffForConnecting =  atoi (P);
	       default :  cerr <<  "Unrecognized option +" <<  thisletter << endl; break;
	       }	    }
	 break;
	 
       default :
	 {
	   if(!gotSeqFile) 
	     {
	       n = strlen(argv[i]);
	       if (n > 100) {cerr << "\n Warning: filename too long max = 100 char \n"; exit(1);}
	       strncpy(optSeqFile,argv[i], 100);
	       gotSeqFile = true;
	     } 
	   else { cerr <<  "Unrecognized option: " <<  argv [i]  << endl ;}
	 }
       }
   }          

 if (verbose) {
   switch(modus)
     {
     case 1: cerr <<  "set option G: generate sequence"<< endl;  break;
     case 2: cerr <<  "set option B: Baum-Welch training"<< endl;  break;
     case 3: cerr <<  "set option D: Posterior Decoding"<< endl;  break;
     case 4: cerr <<  "set option Q: Viterbi Probability and Decoding"<< endl;  break;
     case 5: cerr <<  "set option V: Viterbi Learning"<< endl;  break;
     case 0: cerr <<  "set option P: FB-Probability"<< endl;  break;
     }
 }

 if ((CutoffForWriting >= 0) && (CutoffForWriting <= 1)  )   
   { 
   if (verbose) cerr << "\nSet CutoffForWriting  to " <<  CutoffForWriting ; 
   }
 else 
 {
   CutoffForWriting = kdefCutoff; 
   if (verbose) cerr << "\nUsing default CutoffForWriting = " << CutoffForWriting ; 
 }

 if(!gotModelFile) 
   {
     if (verbose)	 cerr << "\nNo model file was specified, will use default: INPUT/R1.model \n";
     ModelFile = "INPUT/R1.model";
   }
 else {ModelFile = optModelFile;  }	

 if ((!gotSeqFile) & (modus != 1))
   {
     if (verbose)	 cerr << "\nNo sequence file was specified, will use default: SEQUENCES/seqFile \n";
     seqFile = "SEQUENCES/seqFile";
   }
 else { seqFile = optSeqFile;  }	

 if ( (bdsDNAjointly) && (modus!=2) && (modus!=5) ) 
   {
     cerr << "\n Sorry: the DNA double strand option -e is only meaningful in Learning \n";
     exit (1);
   }

 if ((bdsDNA) && (modus==1)) 
   {
     cerr << "\n Sorry: the DNA double strand option -u is not available for Generating Sequences \n";
     exit (1);
   }

 if (DomainThreshold != -1.0) {thresholdNb = 1; thresholds[0]= DomainThreshold;}
 else
   {
     thresholdNb = 5;
     thresholds[0]= kParsingthreshold1;
     thresholds[1]= kParsingthreshold2;
     thresholds[2]= kParsingthreshold3;
     thresholds[3]= kParsingthreshold4;
     thresholds[4]= kParsingthreshold5;
   }

 PlatformMain( ModelFile, seqFile, modus, Number);

cout << "\n";
 return 0;
}
/* MAIN
************************************/
////////////////////////////////////////////////////////////////////////////////////////////////////
